from __future__ import annotations

from typing import Union

from tidy3d.plugins.smatrix.data.modal import ModalComponentModelerData
from tidy3d.plugins.smatrix.data.terminal import TerminalComponentModelerData

ComponentModelerDataType = Union[TerminalComponentModelerData, ModalComponentModelerData]



"""
This module provides integration between Tidy3D and KLayout's Design Rule Check (DRC) engine, allowing you to perform design rule checks on GDS files and Tidy3D objects.

For more information, please see the README.md file in this directory.
For a quickstart example, please see `this notebook <link/to/notebook>`_.
"""

from __future__ import annotations

from .drc import (
    DRCConfig,
    DRCRunner,
    run_drc_on_gds,
)
from .results import DRCResults

__all__ = [
    "DRCConfig",
    "DRCResults",
    "DRCRunner",
    "run_drc_on_gds",
]



"""Defines placements for boundary conditions."""

from __future__ import annotations

from abc import ABC
from typing import Union

import pydantic.v1 as pd

from tidy3d.exceptions import SetupError

from .base import Tidy3dBaseModel
from .types import BoxSurface


class AbstractBCPlacement(ABC, Tidy3dBaseModel):
    """Abstract placement for boundary conditions."""


class StructureBoundary(AbstractBCPlacement):
    """Placement of boundary conditions on the structure's boundary.

    Example
    -------
    >>> bc_placement = StructureBoundary(structure="box")
    """

    structure: str = pd.Field(
        title="Structure Name",
        description="Name of the structure.",
    )


class StructureStructureInterface(AbstractBCPlacement):
    """Placement of boundary conditions between two structures.

    Example
    -------
    >>> bc_placement = StructureStructureInterface(structures=["box", "sphere"])
    """

    structures: tuple[str, str] = pd.Field(
        title="Structures",
        description="Names of two structures.",
    )

    @pd.validator("structures", always=True)
    def unique_names(cls, val):
        """Error if the same structure is provided twice"""
        if val[0] == val[1]:
            raise SetupError(
                "The same structure is provided twice in 'StructureStructureInterface'."
            )
        return val


class MediumMediumInterface(AbstractBCPlacement):
    """Placement of boundary conditions between two mediums.

    Example
    -------
    >>> bc_placement = MediumMediumInterface(mediums=["dieletric", "metal"])
    """

    mediums: tuple[str, str] = pd.Field(
        title="Mediums",
        description="Names of two mediums.",
    )

    @pd.validator("mediums", always=True)
    def unique_names(cls, val):
        """Error if the same structure is provided twice"""
        if val[0] == val[1]:
            raise SetupError("The same medium is provided twice in 'MediumMediumInterface'.")
        return val


class SimulationBoundary(AbstractBCPlacement):
    """Placement of boundary conditions on the simulation box boundary.

    Example
    -------
    >>> bc_placement = SimulationBoundary(surfaces=["x-", "x+"])
    """

    surfaces: tuple[BoxSurface, ...] = pd.Field(
        ("x-", "x+", "y-", "y+", "z-", "z+"),
        title="Surfaces",
        description="Surfaces of simulation domain where to apply boundary conditions.",
    )


class StructureSimulationBoundary(AbstractBCPlacement):
    """Placement of boundary conditions on the simulation box boundary covered by the structure.

    Example
    -------
    >>> bc_placement = StructureSimulationBoundary(structure="box", surfaces=["y-", "y+"])
    """

    structure: str = pd.Field(
        title="Structure Name",
        description="Name of the structure.",
    )

    surfaces: tuple[BoxSurface, ...] = pd.Field(
        ("x-", "x+", "y-", "y+", "z-", "z+"),
        title="Surfaces",
        description="Surfaces of simulation domain where to apply boundary conditions.",
    )


BCPlacementType = Union[
    StructureBoundary,
    StructureStructureInterface,
    MediumMediumInterface,
    SimulationBoundary,
    StructureSimulationBoundary,
]



from __future__ import annotations

from typing import Union

import numpy as np
import pydantic.v1 as pd

from tidy3d.components.base import Tidy3dBaseModel
from tidy3d.components.data.data_array import SpatialDataArray
from tidy3d.constants import PERCMCUBE, SECOND


class FossumCarrierLifetime(Tidy3dBaseModel):
    """
    Parameters for the Fossum carrier lifetime model

    Notes
    -----

        This model expresses the carrier lifetime as a function of the temperature and doping concentration.

        .. math::

            \\tau = \\frac{\\tau_{300} \\left( T/300 \\right)^\\alpha_T}{A + B (N/N_0) + C (N/N_0)^\\alpha}

    Example
    -------
        >>> import tidy3d as td
        >>> default_Si = td.FossumCarrierLifetime(
        ...   tau_300=3.3e-6,
        ...   alpha_T=-0.5,
        ...   N0=7.1e15,
        ...   A=1,
        ...   B=0,
        ...   C=1,
        ...   alpha=1
        ... )

    References
    ----------

        Fossum, J. G., and D. S. Lee. "A physical model for the dependence of carrier lifetime on doping density in nondegenerate silicon." Solid-State Electronics 25.8 (1982): 741-747.

    """

    tau_300: pd.PositiveFloat = pd.Field(
        ..., title="Tau at 300K", description="Carrier lifetime at 300K", units=SECOND
    )

    alpha_T: float = pd.Field(
        ..., title="Exponent for thermal dependence", description="Exponent for thermal dependence"
    )

    N0: pd.PositiveFloat = pd.Field(
        ..., title="Reference concentration", description="Reference concentration", units=PERCMCUBE
    )

    A: float = pd.Field(..., title="Constant A", description="Constant A")

    B: float = pd.Field(..., title="Constant B", description="Constant B")

    C: float = pd.Field(..., title="Constant C", description="Constant C")

    alpha: float = pd.Field(..., title="Exponent constant", description="Exponent constant")


CarrierLifetimeType = Union[FossumCarrierLifetime]


class AugerRecombination(Tidy3dBaseModel):
    """
    Parameters for the Auger recombination model.

    Notes
    -----

        The Auger recombination rate ``R_A`` is primarily defined by the electrons and holes Auger recombination
        coefficients, :math:`C_n` and :math:`C_p`, respectively.

        .. math::

            R_A = \\left( C_n n + C_p p \\right) \\left( np - n_0 p_0 \\right)

    Example
    -------
        >>> import tidy3d as td
        >>> default_Si = td.AugerRecombination(
        ...   c_n=2.8e-31,
        ...   c_p=9.9e-32,
        ... )
    """

    c_n: pd.PositiveFloat = pd.Field(
        ..., title="Constant for electrons", description="Constant for electrons in cm^6/s"
    )

    c_p: pd.PositiveFloat = pd.Field(
        ..., title="Constant for holes", description="Constant for holes in cm^6/s"
    )


class RadiativeRecombination(Tidy3dBaseModel):
    """
    Defines the parameters for the radiative recombination model.

    Notes
    -----

        This is a direct recombination model primarily defined by a radiative recombination coefficient :math:`R_{\\text{rad}}`.

        .. math::

            R_{\\text{rad}} = C \\left( np - n_0 p_0 \\right)

    Example
    -------
        >>> import tidy3d as td
        >>> default_Si = td.RadiativeRecombination(
        ...   r_const=1.6e-14
        ... )
    """

    r_const: float = pd.Field(
        ...,
        title="Radiation constant in cm^3/s",
        description="Radiation constant in cm^3/s",
    )


class ShockleyReedHallRecombination(Tidy3dBaseModel):
    """Defines the parameters for the Shockley-Reed-Hall (SRH) recombination model.

    Notes
    -----

        The recombination rate parameter from this model is defined from [1]_ as follows:

        .. math::

           R_{SRH} = \\frac{n p - n_0 p_0}{\\tau_p \\left(n + \\sqrt{n_0 p_0}\\right) + \\tau_n \\left(p + \\sqrt{n_0 p_0}\\right)}.

        Note that the electron and holes densities are defined within the :class:`SemiconductorMedium`. The electron
        lifetime :math:`\\tau_n` and hole lifetimes :math:`\\tau_p` need to be defined.


        .. [1] Schenk. A model for the field and temperature dependence of shockley-read-hall
               lifetimes in silicon. Solid-State Electronics, 35:1585–1596, 1992.

    Example
    -------
        >>> import tidy3d as td
        >>> default_Si = td.ShockleyReedHallRecombination(
        ...   tau_n=3.3e-6,
        ...   tau_p=4e-6,
        ... )

    Note
    ----
    Important considerations when using this model:

    - This model represents mid-gap traps Shockley-Reed-Hall recombination.
    """

    tau_n: Union[pd.PositiveFloat, CarrierLifetimeType] = pd.Field(
        ..., title="Electron lifetime", description="Electron lifetime", units=SECOND
    )

    tau_p: Union[pd.PositiveFloat, CarrierLifetimeType] = pd.Field(
        ..., title="Hole lifetime", description="Hole lifetime", units=SECOND
    )


class DistributedGeneration(Tidy3dBaseModel):
    """Class that allows to add a distributed generation model.

    Example
    -------
    >>> import tidy3d as td
    >>> import numpy as np
    >>> x = [1,2]
    >>> y = [2,3,4]
    >>> z = [3,4,5,6]
    >>> coords = dict(x=x, y=y, z=z)
    >>> fd = td.SpatialDataArray(np.random.random((2,3,4)), coords=coords)
    >>> dist_g = td.DistributedGeneration(rate=fd)
    """

    rate: SpatialDataArray = pd.Field(
        ...,
        title="Generation rate",
        description="Spatially varying generation rate in cm^-3 s^-1",
        units="cm^-3 s^-1",
    )

    @classmethod
    def from_rate_um3(cls, gen_um3: SpatialDataArray) -> DistributedGeneration:
        """Creates a DistributedGeneration from a SpatialDataArray in um^-3 s^-1."""
        gen_cm3 = np.array(gen_um3.data) * 1e12  # Convert from um^-3 to cm^-3
        new_gen = SpatialDataArray(gen_cm3, coords=gen_um3.coords)
        return cls(rate=new_gen)

    @pd.root_validator(skip_on_failure=True)
    def check_spatialdataarray_dimensions(cls, values):
        """Check that the SpatialDataArray is at least 2D:"""

        rate = values.get("rate")

        zero_dims = [d for d in ["x", "y", "z"] if len(rate.coords[d]) <= 1]

        if len(zero_dims) > 1:
            raise ValueError("SpatialDataArray must be at least 2D.")

        return values



"""Methods for integrating KLayout's DRC with Tidy3D."""

from __future__ import annotations

import re
from pathlib import Path
from subprocess import run
from typing import Union

import pydantic.v1 as pd
from pydantic.v1 import validator

from tidy3d.components.base import Tidy3dBaseModel
from tidy3d.components.geometry.base import Geometry
from tidy3d.components.simulation import Simulation
from tidy3d.components.structure import Structure
from tidy3d.exceptions import ValidationError
from tidy3d.log import get_logging_console
from tidy3d.plugins.klayout.drc.defaults import (
    DEFAULT_GDSFILE,
    DEFAULT_RESULTSFILE,
    DEFAULT_VERBOSE,
)
from tidy3d.plugins.klayout.drc.results import DRCResults
from tidy3d.plugins.klayout.util import check_installation


class DRCConfig(Tidy3dBaseModel):
    """Configuration for KLayout DRC."""

    gdsfile: pd.FilePath = pd.Field(
        title="GDS File",
        description="The path to the GDS file to write the Tidy3D object to.",
    )
    drc_runset: pd.FilePath = pd.Field(
        title="DRC Runset file",
        description="Path to the KLayout DRC runset file.",
    )
    resultsfile: Path = pd.Field(
        title="DRC Results File",
        description="Path to the KLayout DRC results file.",
    )
    verbose: bool = pd.Field(
        title="Verbose",
        description="Whether to print logging.",
    )

    @validator("gdsfile")
    def _validate_gdsfile_filetype(cls, v: pd.FilePath) -> pd.FilePath:
        """Check GDS filetype is ``.gds``."""
        if v.suffix != ".gds":
            raise ValidationError(f"GDS file '{v}' must end with '.gds'.")
        return v

    @validator("drc_runset")
    def _validate_drc_runset_filetype(cls, v: pd.FilePath) -> pd.FilePath:
        """Check DRC runset filetype is ``.drc``."""
        if v.suffix != ".drc":
            raise ValidationError(f"DRC runset file '{v}' must end with '.drc'.")
        return v

    @validator("drc_runset")
    def _validate_drc_runset_format(cls, v: pd.FilePath) -> pd.FilePath:
        """Check if the DRC runset file is formatted correctly.
        The checks are:
        1. The GDS source must be loaded with 'source($gdsfile)'.
        2. The report must be defined as 'report("<your string>", $resultsfile)'.
        """
        with Path(v).open("r") as f:
            content = f.read()
            if not re.search(r"source\(\s*\$gdsfile\s*\)", content):
                raise ValidationError(
                    "DRC runset is not formatted correctly. The GDS source must be loaded with 'source($gdsfile)'. Please refer to the documentation at 'tidy3d/plugins/klayout/drc/README.md' for more details."
                )
            if not re.search(r"""report\(['"](.*?)['"],\s*\$resultsfile\)""", content):
                raise ValidationError(
                    "DRC runset is not formatted correctly. The report must be defined as 'report(\"<your report name>\", $resultsfile)'. Please refer to the documentation at 'tidy3d/plugins/klayout/drc/README.md' for more details."
                )
        return v


class DRCRunner(Tidy3dBaseModel):
    """A class for running KLayout DRC. Can be used to run DRC on a Tidy3D object or a GDS file.

    Parameters
    ----------
    drc_runset : Path
        The path to the KLayout DRC runset file.
    verbose : bool
        Whether to print logging. Default is ``True``.

    Example
    -------
    >>> # Running DRC on a GDS file:
    >>> from tidy3d.plugins.klayout.drc import DRCRunner
    >>> runner = DRCRunner(drc_runset="my_drc_runset.drc", verbose=True) # doctest: +SKIP
    >>> results = runner.run(source="my_layout.gds", resultsfile="drc_results.lyrdb") # doctest: +SKIP
    >>> print(results) # doctest: +SKIP
    >>> # Running DRC on a Tidy3D object:
    >>> import tidy3d as td
    >>> from tidy3d.plugins.klayout.drc import DRCRunner
    >>> vertices = [(-2, 0), (-1, 1), (0, 0.5), (1, 1), (2, 0), (0, -1)]
    >>> geom = td.PolySlab(vertices=vertices, slab_bounds=(0, 0.22), axis=2)
    >>> runner = DRCRunner(drc_runset="my_drc_runset.drc", verbose=True) # doctest: +SKIP
    >>> results = runner.run(source=geom, td_object_gds_savefile="geom.gds", resultsfile="drc_results.lyrdb", z=0.1, gds_layer=0, gds_dtype=0) # doctest: +SKIP
    >>> print(results) # doctest: +SKIP
    """

    drc_runset: pd.FilePath = pd.Field(
        title="DRC Runset file",
        description="Path to the KLayout DRC runset file.",
    )
    verbose: bool = pd.Field(
        default=DEFAULT_VERBOSE,
        title="Verbose",
        description="Whether to print logging.",
    )

    def run(
        self,
        source: Union[Geometry, Structure, Simulation, Path],
        td_object_gds_savefile: Path = DEFAULT_GDSFILE,
        resultsfile: Path = DEFAULT_RESULTSFILE,
        **to_gds_file_kwargs,
    ) -> None:
        """Runs KLayout's DRC on a GDS file or a Tidy3D object. The Tidy3D object can be a :class:`.Geometry`, :class:`.Structure`, or :class:`.Simulation`.

        Parameters
        ----------
        source : Union[:class:`.Geometry`, :class:`.Structure`, :class:`.Simulation`, Path]
            The :class:`.Geometry`, :class:`.Structure`, :class:`.Simulation`, or GDS file to run DRC on.
        td_object_gds_savefile : Path
            The path to save the Tidy3D object to. Defaults to ``"layout.gds"``.
        resultsfile : Path
            The path to save the KLayout DRC results file to. Defaults to ``"drc_results.lyrdb"``.
        **to_gds_file_kwargs
            Additional keyword arguments to pass to the Tidy3D object-specific ``to_gds_file()`` method.

        Returns
        -------
        :class:`.DRCResults`
            The DRC results object containing violations and status.

        Example
        -------
        Running DRC on a GDS file:
        >>> from tidy3d.plugins.klayout.drc import DRCRunner
        >>> runner = DRCRunner(drc_runset="my_drc_runset.drc", verbose=True) # doctest: +SKIP
        >>> results = runner.run(source="my_layout.gds") # doctest: +SKIP
        >>> print(results) # doctest: +SKIP

        Running DRC on a Tidy3D object:
        >>> import tidy3d as td
        >>> from tidy3d.plugins.klayout.drc import DRCRunner
        >>> vertices = [(-2, 0), (-1, 1), (0, 0.5), (1, 1), (2, 0), (0, -1)]
        >>> geom = td.PolySlab(vertices=vertices, slab_bounds=(0, 0.22), axis=2)
        >>> runner = DRCRunner(drc_runset="my_drc_runset.drc", verbose=True) # doctest: +SKIP
        >>> results = runner.run(source=geom, z=0.1, gds_layer=0, gds_dtype=0) # doctest: +SKIP
        >>> print(results) # doctest: +SKIP
        """
        if isinstance(source, (Geometry, Structure, Simulation)):
            gdsfile = td_object_gds_savefile
            if self.verbose:
                console = get_logging_console()
                console.log(f"Writing Tidy3D object to GDS file '{gdsfile}'.")
            source.to_gds_file(fname=gdsfile, **to_gds_file_kwargs)
        else:
            gdsfile = source

        config = DRCConfig(
            gdsfile=gdsfile,
            drc_runset=self.drc_runset,
            resultsfile=resultsfile,
            verbose=self.verbose,
        )
        return run_drc_on_gds(config=config)


def run_drc_on_gds(config: DRCConfig) -> DRCResults:
    """Runs KLayout's DRC on a GDS file.

    Parameters
    ----------
    config : :class:`.DRCConfig`
        The configuration for the DRC run.

    Returns
    -------
    :class:`.DRCResults`
        The DRC results object containing violations and status.

    Example
    -------
    >>> from tidy3d.plugins.klayout.drc import run_drc_on_gds, DRCConfig
    >>> config = DRCConfig(gdsfile="geom.gds", drc_runset="my_drc_runset.drc", resultsfile="drc_results.lyrdb", verbose=True) # doctest: +SKIP
    >>> results = run_drc_on_gds(config) # doctest: +SKIP
    >>> print(results) # doctest: +SKIP
    """
    check_installation(raise_error=True)

    if config.verbose:
        console = get_logging_console()
        console.log(
            f"Running KLayout DRC on GDS file '{config.gdsfile}' with runset '{config.drc_runset}' and saving results to '{config.resultsfile}'..."
        )
    # run klayout DRC as a subprocess
    output = run(
        [
            "klayout",
            "-b",
            "-r",
            config.drc_runset,
            "-rd",
            f"gdsfile={config.gdsfile}",
            "-rd",
            f"resultsfile={config.resultsfile}",
        ],
        capture_output=True,
    )

    if output.returncode != 0:
        raise RuntimeError(f"KLayout DRC failed with error message: '{output.stderr}'.")
    if config.verbose:
        console.log("KLayout DRC completed successfully.")

    return DRCResults.load(resultsfile=config.resultsfile)



"""Methods for loading and parsing KLayout DRC results."""

from __future__ import annotations

import re
import xml.etree.ElementTree as ET
from pathlib import Path
from typing import Union

import pydantic.v1 as pd

from tidy3d.components.base import Tidy3dBaseModel, cached_property
from tidy3d.components.types import Coordinate2D
from tidy3d.exceptions import FileError

# Types for DRC markers
DRCEdge = tuple[Coordinate2D, Coordinate2D]
DRCEdgePair = tuple[DRCEdge, DRCEdge]
DRCPolygon = tuple[Coordinate2D, ...]
DRCMultiPolygon = tuple[DRCPolygon, ...]


def parse_edge(value: str) -> EdgeMarker:
    """
    Extract coordinates from edge format: ``(x1,y1;x2,y2)``.

    Parameters
    ----------
    value : str
        The edge value string from DRC result database, with format ``(x1,y1;x2,y2)``.

    Returns
    -------
    :class:`.EdgeMarker`
        :class:`.EdgeMarker` containing start and end points of the edge.

    Raises
    ------
    ValueError
        If the edge format is invalid.
    """
    # Extract coordinates from edge format: (x1,y1;x2,y2)
    pattern = r"\(([\d.-]+),([\d.-]+);([\d.-]+),([\d.-]+)\)"
    match = re.match(pattern, value)
    if match:
        coords = [float(x) for x in match.groups()]
        return EdgeMarker(edge=((coords[0], coords[1]), (coords[2], coords[3])))
    raise ValueError(f"Invalid edge format: '{value}'.")


def parse_edge_pair(value: str) -> EdgePairMarker:
    """
    Extract coordinates from edge-pair format: ``(x1,y1;x2,y2)|(x3,y3;x4,y4)``.

    Parameters
    ----------
    value : str
        The edge-pair value string from DRC result database, with format ``(x1,y1;x2,y2)|(x3,y3;x4,y4)``.

    Returns
    -------
    :class:`.EdgePairMarker`
        :class:`.EdgePairMarker` containing both edges' coordinates.

    Raises
    ------
    ValueError
        If the edge-pair format is invalid.
    """
    # Extract coordinates from edge-pair format: (x1,y1;x2,y2)|(x3,y3;x4,y4)
    pattern = (
        r"\(([\d.-]+),([\d.-]+);([\d.-]+),([\d.-]+)\)\|\(([\d.-]+),([\d.-]+);([\d.-]+),([\d.-]+)\)"
    )
    match = re.match(pattern, value)
    if match:
        coords = [float(x) for x in match.groups()]
        return EdgePairMarker(
            edge_pair=(
                ((coords[0], coords[1]), (coords[2], coords[3])),
                ((coords[4], coords[5]), (coords[6], coords[7])),
            )
        )
    raise ValueError(f"Invalid edge-pair format: '{value}'.")


def parse_polygon_coordinates(coords_str: str) -> DRCPolygon:
    """Parse coordinates for a single polygon into a tuple of (x, y) pairs.

    Parameters
    ----------
    coords_str : str
        The string of coordinates for a single polygon, with format ``x1,y1;x2,y2;...``.

    Returns
    -------
    DRCPolygon
        A tuple of (x, y) pairs for the polygon.

    Raises
    ------
    ValueError
        If the coordinate format is invalid.
    """
    coords_pattern = r"([\d.-]+),([\d.-]+)"  # match the coordinates

    # Split by semicolon to validate each coordinate pair
    coord_pairs = coords_str.split(";")

    # Each coordinate pair should match the pattern exactly
    coords = []
    for pair in coord_pairs:
        match = re.match(coords_pattern, pair)
        if not match:
            raise ValueError(f"Invalid coordinate pair in polygon: '{pair}'.")
        coords.append((float(match.group(1)), float(match.group(2))))

    return tuple(coords)


def parse_polygons(value: str) -> MultiPolygonMarker:
    """
    Extract coordinates from polygon format: ``(x1,y1;x2,y2;...)`` including multiple polygons separated by ``/``.

    Parameters
    ----------
    value : str
        The polygon value string from DRC result database, with format ``(x1,y1;x2,y2;...)``
        or multiple polygons separated by ``/`` like ``(x1,y1;.../x3,y3;...)``.

    Returns
    -------
    :class:`.MultiPolygonMarker`
        :class:`.MultiPolygonMarker` containing one or more polygon shapes.

    Raises
    ------
    ValueError
        If the polygon format is invalid or contains incomplete coordinate pairs.
    """
    # Extract the full content inside outer parentheses
    outer_pattern = r"\((.*)\)"
    match = re.match(outer_pattern, value)
    if not match:
        raise ValueError(f"Invalid polygon format: '{value}'.")

    coords_content = match.group(1)

    # Parse multiple polygons separated by '/'
    polygon_parts = coords_content.split("/")
    polygons = []
    for part in polygon_parts:
        polygons.append(parse_polygon_coordinates(part.strip()))

    return MultiPolygonMarker(polygons=tuple(polygons))


def parse_violation_value(value: str) -> Union[EdgeMarker, EdgePairMarker, MultiPolygonMarker]:
    """
    Parse a violation value based on its type (edge, edge-pair, or polygon).

    Parameters
    ----------
    value : str
        The value string from DRC result database.

    Returns
    -------
    Union[:class:`.EdgeMarker`, :class:`.EdgePairMarker`, :class:`.MultiPolygonMarker`]
        The parsed violation marker.

    Raises
    ------
    ValueError
        If the violation marker type is invalid.
    """
    if value.startswith("edge: "):
        return parse_edge(value=value.replace("edge: ", ""))
    elif value.startswith("edge-pair: "):
        return parse_edge_pair(value=value.replace("edge-pair: ", ""))
    elif value.startswith("polygon: "):
        return parse_polygons(value=value.replace("polygon: ", ""))
    raise ValueError(
        f"Invalid marker type (should start with 'edge:', 'edge-pair:', or 'polygon:'): '{value}'."
    )


class EdgeMarker(Tidy3dBaseModel):
    """A class for storing KLayout DRC edge marker results."""

    edge: DRCEdge = pd.Field(
        title="DRC Edge Marker",
        description="The edge marker of the DRC violation. The format is ((x1, y1), (x2, y2)).",
    )


class EdgePairMarker(Tidy3dBaseModel):
    """A class for storing KLayout DRC edge pair marker results."""

    edge_pair: DRCEdgePair = pd.Field(
        title="DRC Edge Pair Marker",
        description="The edge pair marker of the DRC violation. The format is (edge1, edge2), where an edge has format ((x1, y1), (x2, y2)).",
    )


class MultiPolygonMarker(Tidy3dBaseModel):
    """A class for storing KLayout DRC multi-polygon marker results."""

    polygons: DRCMultiPolygon = pd.Field(
        title="DRC Multi-Polygon Marker",
        description="The multi-polygon marker of the DRC violation. The format is (polygon1, polygon2, ...), where each polygon has format ((x1, y1), (x2, y2), ...).",
    )


DRCMarker = Union[EdgeMarker, EdgePairMarker, MultiPolygonMarker]


class DRCViolation(Tidy3dBaseModel):
    """A class for storing KLayout DRC violation results for a single category."""

    category: str = pd.Field(
        title="DRC Violation Category", description="The category of the DRC violation."
    )
    markers: tuple[DRCMarker, ...] = pd.Field(
        title="DRC Markers", description="Tuple of DRC markers in this category."
    )

    @cached_property
    def count(self) -> int:
        """The number of DRC markers in this category."""
        return len(self.markers)

    def __str__(self) -> str:
        """Get a nice string summary of the number of markers in this category."""
        return f"{self.category}: {self.count}"


class DRCResults(Tidy3dBaseModel):
    """A class for loading and storing KLayout DRC results."""

    violations_by_category: dict[str, DRCViolation] = pd.Field(
        title="DRC Violations", description="Dictionary of DRC violations by category."
    )

    @cached_property
    def is_clean(self) -> bool:
        """Whether the DRC is clean (no violations)."""
        return all(v.count == 0 for v in self.violations_by_category.values())

    @cached_property
    def violation_counts(self) -> dict[str, int]:
        """Counts violations by category.

        Returns
        -------
        dict[str, int]
            A dictionary of violation counts for each category.
        """
        return {
            category: violation.count for category, violation in self.violations_by_category.items()
        }

    @cached_property
    def categories(self) -> tuple[str, ...]:
        """A tuple of all DRC categories."""
        return tuple(self.violations_by_category.keys())

    def __getitem__(self, category: str) -> DRCViolation:
        """Get DRC violation result by category.

        Parameters
        ----------
        category : str
            The category of the DRC violation.

        Returns
        -------
        :class:`.DRCViolation`
            The DRC violation result for the given category.
        """
        return self.violations_by_category[category]

    def __str__(self) -> str:
        """Get a nice string representation of the DRC results."""
        summary = "DRC results summary\n"
        summary += "--------------------------------\n"
        summary += f"Total violations: {sum(violation.count for violation in self.violations_by_category.values())}\n\n"
        summary += "Violations by category:\n"
        for violation in self.violations_by_category.values():
            summary += violation.__str__() + "\n"
        return summary

    @classmethod
    def load(cls, resultsfile: Union[str, Path]) -> DRCResults:
        """Create a :class:`.DRCResults` instance from a results file.

        Parameters
        ----------
        resultsfile : Union[str, Path]
            Path to the KLayout DRC results file.

        Returns
        -------
        :class:`.DRCResults`
            A :class:`.DRCResults` object containing the DRC results.

        Raises
        ------
        FileNotFoundError
            If the DRC result file is not found.
        ET.ParseError
            If the DRC result file is not a valid XML file.

        Example
        -------
        >>> from tidy3d.plugins.klayout.drc import DRCResults
        >>> results = DRCResults.load(resultsfile="drc_results.lyrdb") # doctest: +SKIP
        >>> print(results) # doctest: +SKIP
        """
        return cls(violations_by_category=violations_from_file(resultsfile=resultsfile))


def violations_from_file(resultsfile: Union[str, Path]) -> dict[str, DRCViolation]:
    """Loads a KLayout DRC results file and returns the results as a dictionary of :class:`.DRCViolation` objects.

    Parameters
    ----------
    resultsfile : Union[str, Path]
        Path to the KLayout DRC results file.

    Returns
    -------
    dict[str, :class:`.DRCViolation`]
        A dictionary of :class:`.DRCViolation` objects for each category.

    Raises
    ------
    FileNotFoundError
        If the DRC result file is not found.
    ET.ParseError
        If the DRC result file is not a valid XML file.
    """
    # Parse the results file
    try:
        xmltree = ET.parse(resultsfile)
    except FileNotFoundError as err:
        raise FileError(f"DRC result file not found: '{resultsfile}'.") from err
    except ET.ParseError as err:
        raise ET.ParseError(f"Invalid XML format in DRC result file: '{resultsfile}'.") from err

    # Initialize violations dict with all the categories
    violations = {}
    for category in xmltree.getroot().findall(".//categories/category/name"):
        violations[category.text] = DRCViolation(category=category.text, markers=())

    # Parse markers
    for item in xmltree.getroot().findall(".//item"):
        category = item.find("category").text
        value = item.find("values/value").text
        marker = parse_violation_value(value)
        violations[category] = DRCViolation(
            category=category,
            markers=(*violations[category].markers, marker),
        )
    return violations



"""Classes for creating data based on analytic beams like plane wave, Gaussian beam, and
astigmatic Gaussian beam."""

from __future__ import annotations

from abc import abstractmethod
from typing import Literal, Optional, Union

import autograd.numpy as np
import pydantic.v1 as pd

from tidy3d.constants import C_0, ETA_0, HERTZ, MICROMETER, RADIAN

from .base import cached_property
from .data.data_array import ScalarFieldDataArray
from .data.monitor_data import FieldData
from .geometry.base import Box
from .grid.grid import Coords, Grid
from .medium import Medium, MediumType
from .monitor import FieldMonitor
from .source.field import FixedAngleSpec, FixedInPlaneKSpec
from .types import TYPE_TAG_STR, Direction, FreqArray, Numpy
from .validators import assert_plane

DEFAULT_RESOLUTION = 200


class BeamProfile(Box):
    """Base class for handling analytic beams."""

    resolution: float = pd.Field(
        DEFAULT_RESOLUTION,
        title="Sampling resolution",
        description="Sampling resolution in the tangential directions of the beam (defines a "
        "number of equally spaced points).",
        units=MICROMETER,
    )

    freqs: FreqArray = pd.Field(
        ...,
        title="Frequencies",
        description="List of frequencies at which the beam is sampled.",
        units=HERTZ,
    )

    background_medium: MediumType = pd.Field(
        Medium(),
        title="Background Medium",
        description="Background medium in which the beam is embedded.",
    )

    angle_theta: float = pd.Field(
        0.0,
        title="Polar Angle",
        description="Polar angle of the propagation axis from the normal axis.",
        units=RADIAN,
    )

    angle_phi: float = pd.Field(
        0.0,
        title="Azimuth Angle",
        description="Azimuth angle of the propagation axis in the plane orthogonal to the "
        "normal axis.",
        units=RADIAN,
    )

    pol_angle: float = pd.Field(
        0.0,
        title="Polarization Angle",
        description="Specifies the angle between the electric field polarization of the "
        "beam and the plane defined by the normal axis and the propagation axis (rad). "
        "``pol_angle=0`` (default) specifies P polarization, "
        "while ``pol_angle=np.pi/2`` specifies S polarization. "
        "At normal incidence when S and P are undefined, ``pol_angle=0`` defines: "
        "- ``Ey`` polarization for propagation along ``x``."
        "- ``Ex`` polarization for propagation along ``y``."
        "- ``Ex`` polarization for propagation along ``z``.",
        units=RADIAN,
    )

    direction: Direction = pd.Field(
        "+",
        title="Direction",
        description="Specifies propagation in the positive or negative direction of the normal "
        "axis.",
    )

    _plane_validator = assert_plane()

    @property
    def grid(self) -> Grid:
        """Return a Grid object on which the beam will be sampled."""
        dim_n, dim_tan = self.pop_axis("xyz", self.size.index(0.0))
        bounds_n, bounds_tan = self.pop_axis(np.array(self.bounds).T, self.size.index(0.0))
        boundaries = {
            dim_n: bounds_n,
            dim_tan[0]: np.linspace(bounds_tan[0][0], bounds_tan[0][1], int(self.resolution)),
            dim_tan[1]: np.linspace(bounds_tan[1][0], bounds_tan[1][1], int(self.resolution)),
        }
        return Grid(boundaries=boundaries)

    @property
    def monitor(self) -> FieldMonitor:
        """``FieldMonitor`` with the same center, size, and frequencies as the beam, to be used
        in the ``FieldData`` created by the ``field_data`` method."""
        return FieldMonitor(
            size=self.size,
            center=self.center,
            freqs=self.freqs,
            name="<<BEAM_DATA>>",
        )

    @cached_property
    def field_data(self) -> FieldData:
        """Compute a FieldData for the spatial E and H field amplitudes of an analytically defined
        beam."""

        background_n = np.zeros(len(self.freqs), dtype=complex)
        for freq_id, freq in enumerate(self.freqs):
            eps = self.background_medium.eps_model(freq)
            nk_medium = self.background_medium.eps_complex_to_nk(eps)
            background_n[freq_id] = np.squeeze(nk_medium[0]) + 1j * np.squeeze(nk_medium[1])

        data_dict = self._field_data_on_grid(grid=self.grid, background_n=background_n)
        data_raw = FieldData(monitor=self.monitor, grid_expanded=self.grid, **data_dict)

        # Normalize by the flux
        fields_norm = {}
        flux = np.abs(data_raw.flux)
        for field_name, field_data in data_raw.field_components.items():
            fields_norm[field_name] = (field_data / np.sqrt(flux)).astype(field_data.dtype)

        return data_raw.updated_copy(**fields_norm)

    def _field_data_on_grid(self, grid: Grid, background_n: Numpy, colocate=True) -> dict:
        """Compute the field data for each field component on a grid for the beam.
        A dictionary of the scalar field data arrays is returned, not yet packaged as ``FieldData``.
        """

        field_components = ["Ex", "Ey", "Ez", "Hx", "Hy", "Hz"]
        if colocate:
            # Just sample at the grid boundaries for each field component
            # We skip the last boundary because that's how things also work in the solver (no
            # field defined at the last boundary)
            coords_dict = {key: val[:-1] for key, val in grid.boundaries.to_dict.items()}
            grid_dict = {component: Coords(**coords_dict) for component in field_components}
        else:
            # Yee grid for each field component
            grid_dict = grid.yee.grid_dict

        # Compute each field component over the beam grid
        scalar_fields = {}
        for comp, field in enumerate(field_components):
            x, y, z = grid_dict[field].to_list
            # Center component grid at beam center
            x_c, y_c, z_c = (coords - cent for coords, cent in zip((x, y, z), self.center))
            # Stack E and H fields
            field_vals = self.analytic_beam(x_c, y_c, z_c, background_n, field=field[0])
            # Get the current field component
            field_vals = field_vals[comp % 3]
            # Make the ScalarFieldDataArray for the current component
            coords = {"x": x, "y": y, "z": z, "f": np.array(self.freqs)}
            field_data = ScalarFieldDataArray(field_vals, coords=coords)
            scalar_fields[field] = field_data

        return scalar_fields

    @abstractmethod
    def scalar_field(self, points: Numpy, background_n: float) -> Numpy:
        """Scalar field corresponding to the analytic beam in coordinate system such that the
        propagation direction is z and the ``E``-field is entirely ``x``-polarized. The field is
        computed on an unstructured array ``points`` of shape ``(3, ...)``."""

    def analytic_beam_z_normal(
        self, points: Numpy, background_n: float, field: Literal["E", "H"]
    ) -> Numpy:
        """Analytic beam with all the beam parameters but assuming ``z`` as the normal axis."""

        # Add a frequency dimension to points
        points = np.repeat(points[:, :, np.newaxis], len(self.freqs), axis=2)

        # Rotate points to axes where propagation is along z
        points_prop_z = self._rotate_points_z(points, background_n)

        # Reflection at the z = 0 plane for negative direction
        if self.direction == "-":
            points_prop_z[2, ...] *= -1

        # Get fields polarized along x and propagating along z
        scalar_field = self.scalar_field(points_prop_z, background_n)

        # Set the correct field component based on the scalar field values
        field_vals = np.zeros(points.shape, dtype=np.complex128)
        if field == "E":
            field_vals[0, ...] = scalar_field
        else:
            field_vals[1, ...] = scalar_field / ETA_0 * background_n

        # Rotate polarization
        field_vals = self.rotate_points(field_vals, [0, 0, 1], self.pol_angle)

        # Reflection of the field components at the z = 0 plane for negative direction
        if self.direction == "-":
            if field == "E":
                field_vals[2, :] *= -1
            else:
                field_vals[:2, :] *= -1

        # Rotate the fields back to the original propagation axes
        field_vals = self._inverse_rotate_field_vals_z(field_vals, background_n)

        return field_vals

    def analytic_beam(
        self,
        x: Numpy,
        y: Numpy,
        z: Numpy,
        background_n: float,
        field: Literal["E", "H"],
    ) -> Numpy:
        """Sample the analytic beam fields on a cartesian grid of points in x, y, z."""

        # Make a meshgrid
        (x_mesh, y_mesh, z_mesh) = np.meshgrid(x, y, z, indexing="ij")
        (Nx, Ny, Nz) = x_mesh.shape

        # Move to coordinates where the normal axis is z
        (z_new, (x_new, y_new)) = self.pop_axis((x_mesh, y_mesh, z_mesh), axis=self.size.index(0.0))
        points = np.stack([coords.ravel() for coords in (x_new, y_new, z_new)], axis=0)

        # Sample the plane wave fields
        f_x, f_y, f_z = self.analytic_beam_z_normal(points, background_n, field)

        # Move back to original coordinates
        field_vals = np.stack(self.unpop_axis(f_z, (f_x, f_y), axis=self.size.index(0.0)), axis=0)

        # H field gets a -1 factor if a reflection was involved
        if self.size.index(0.0) == 1 and field == "H":
            field_vals *= -1

        # Reshape to (3, Nx, Ny, Nz, num_freqs)
        return np.reshape(field_vals, (3, Nx, Ny, Nz, len(self.freqs)))

    def _rotate_points_z(self, points: Numpy, background_n: Numpy) -> Numpy:
        """Rotate points to new coordinates where z is the propagation axis."""
        points_prop_z = self.rotate_points(points, [0, 0, 1], -self.angle_phi)
        points_prop_z = self.rotate_points(points_prop_z, [0, 1, 0], -self.angle_theta)
        return points_prop_z

    def _inverse_rotate_field_vals_z(self, field_vals: Numpy, background_n: Numpy) -> Numpy:
        """Rotate field values from coordinates where z is the propagation axis to angled
        coordinates."""
        field_vals = self.rotate_points(field_vals, [0, 1, 0], self.angle_theta)
        field_vals = self.rotate_points(field_vals, [0, 0, 1], self.angle_phi)
        return field_vals


class PlaneWaveBeamProfile(BeamProfile):
    """Component for constructing plane wave beam data. The normal direction is implicitly
    defined by the ``size`` parameter.

    See also :class:`.PlaneWave`.
    """

    angular_spec: Union[FixedInPlaneKSpec, FixedAngleSpec] = pd.Field(
        FixedAngleSpec(),
        title="Angular Dependence Specification",
        description="Specification of plane wave propagation direction dependence on wavelength.",
        discriminator=TYPE_TAG_STR,
    )

    as_fixed_angle_source: bool = pd.Field(
        False,
        title="Fixed Angle Flag",
        description="Fixed angle flag. Only used internally when computing source beams for "
        "injection in an FDTD simulation with fixed angle boudnaries. Use ``angular_spec`` to "
        "switch between waves with fixed angle and fixed in-plane k.",
    )

    angle_theta_frequency: Optional[float] = pd.Field(
        None,
        title="Frequency at Which Angle Theta is Defined",
        description="Frequency for which ``angle_theta`` is set. This only has an effect for "
        "fixed in-plane wave-vector beams. If not supplied, the average of the beam ``freqs`` is "
        "used.",
    )

    @property
    def _angle_theta_frequency(self):
        if not self.angle_theta_frequency:
            return np.mean(self.freqs)
        return self.angle_theta_frequency

    def in_plane_k(self, background_n: float):
        """In-plane wave vector. Only the real part is taken so the beam has no in-plane decay."""
        k0 = 2 * np.pi * self._angle_theta_frequency / C_0 * background_n
        k_in_plane = k0.real * np.sin(self.angle_theta)
        return [k_in_plane * np.cos(self.angle_phi), k_in_plane * np.sin(self.angle_phi)]

    def scalar_field(self, points: Numpy, background_n: float) -> Numpy:
        """Scalar field for plane wave.
        Scalar field corresponding to the analytic beam in coordinate system such that the
        propagation direction is z and the ``E``-field is entirely ``x``-polarized. The field is
        computed on an unstructured array ``points`` of shape ``(3, N_points, N_freqs)``.
        For the special case of fixed in-plane k, the propagation axis is different at every
        frequency, and the points a frquency-dependent rotation has been applied to the
        ``points`` in ``self._rotate_points_z``.
        """

        kz = 2 * np.pi * np.array(self.freqs) / C_0 * background_n
        if self.as_fixed_angle_source:
            kz *= np.cos(self.angle_theta)
        return np.exp(1j * points[2] * kz)

    def _angle_theta_actual(self, background_n: Numpy) -> Numpy:
        """Compute the frequency-dependent actual propagation angle theta."""
        k0 = 2 * np.pi * np.array(self.freqs) / C_0 * background_n
        kx, ky = self.in_plane_k(background_n)
        k_perp = np.sqrt(kx**2 + ky**2)
        return np.real(np.arcsin(k_perp / k0)) * np.sign(self.angle_theta)

    def _rotate_points_z(self, points: Numpy, background_n: Numpy) -> Numpy:
        """Rotate points to new coordinates where z is the propagation axis."""
        if self.as_fixed_angle_source:
            # For fixed-angle, we do not rotate the points
            return points
        if isinstance(self.angular_spec, FixedInPlaneKSpec):
            # For fixed in-plane k, the rotation is angle-dependent
            points = self.rotate_points(points, [0, 0, 1], -self.angle_phi)
            angle_theta_actual = self._angle_theta_actual(background_n=background_n)
            for ind, theta_actual in enumerate(angle_theta_actual):
                points[:, :, ind] = self.rotate_points(points[:, :, ind], [0, 1, 0], -theta_actual)
            return points
        return super()._rotate_points_z(points, background_n)

    def _inverse_rotate_field_vals_z(self, field_vals: Numpy, background_n: Numpy) -> Numpy:
        """Rotate field values from coordinates where z is the propagation axis to angled
        coordinates. Special handling is needed if fixed in-plane k wave."""
        if isinstance(self.angular_spec, FixedInPlaneKSpec):
            # For fixed in-plane k, the rotation is angle-dependent
            angle_theta_actual = self._angle_theta_actual(background_n=background_n)
            for ind, theta_actual in enumerate(angle_theta_actual):
                field_vals[:, :, ind] = self.rotate_points(
                    field_vals[:, :, ind], [0, 1, 0], theta_actual
                )
            field_vals = self.rotate_points(field_vals, [0, 0, 1], self.angle_phi)
            return field_vals
        return super()._inverse_rotate_field_vals_z(field_vals, background_n)


class GaussianBeamProfile(BeamProfile):
    """Component for constructing Gaussian beam data. The normal direction is implicitly
    defined by the ``size`` parameter.

    See also :class:`.GaussianBeam`.
    """

    waist_radius: pd.PositiveFloat = pd.Field(
        1.0,
        title="Waist Radius",
        description="Radius of the beam at the waist.",
        units=MICROMETER,
    )

    waist_distance: float = pd.Field(
        0.0,
        title="Waist Distance",
        description="Distance from the beam waist along the propagation direction. "
        "A positive value means the waist is positioned behind the beam, considering the propagation direction. "
        "For example, for a beam propagating in the ``+`` direction, a positive value of ``beam_distance`` "
        "means the beam waist is positioned in the ``-`` direction (behind the beam). "
        "A negative value means the beam waist is in the ``+`` direction (in front of the beam). "
        "For an angled beam, the distance is defined along the rotated propagation direction.",
        units=MICROMETER,
    )

    def beam_params(self, z: Numpy, k0: Numpy) -> tuple[Numpy, Numpy, Numpy]:
        """Compute the parameters needed to evaluate a Gaussian beam at z.

        Parameters
        ----------
        z : Numpy
            Axial distance from the beam center.
        k0 : Numpy
            Wave vector magnitude.
        """

        w_0, z_0 = self.waist_radius, self.waist_distance
        z_r = w_0**2 * k0 / 2  # shape k0
        w_z = w_0 * np.sqrt(1 + ((z + z_0) / z_r) ** 2)  # shape (Np, Nk0)
        # inv_r_z shape (Np, Nk0)
        inv_r_z = (z + z_0) / ((z + z_0) ** 2 + z_r**2)
        # we choose gauge such that psi_g = 0 at z = 0 (beam plane)
        # this is needed for a proper interpolation between different frequencies
        # psi_g shape (Np, Nk0)
        psi_g = np.arctan((z + z_0) / z_r) - np.arctan(z_0 / z_r)
        return w_z, inv_r_z, psi_g

    def scalar_field(self, points: Numpy, background_n: float) -> Numpy:
        """Scalar field for Gaussian beam.
        Scalar field corresponding to the analytic beam in coordinate system such that the
        propagation direction is z and the ``E``-field is entirely ``x``-polarized. The field is
        computed on an unstructured array ``points`` of shape ``(3, ...)``.
        """
        k0 = 2 * np.pi * np.array(self.freqs) / C_0 * background_n
        x, y, z = points
        w_0 = self.waist_radius
        w_z, inv_r_z, psi_g = self.beam_params(z, k0)
        r_2 = x**2 + y**2
        scalar_gaussian = w_0 / w_z
        scalar_gaussian *= np.exp(-r_2 / w_z**2)
        scalar_gaussian *= np.exp(1j * (z * k0 + r_2 * k0 / 2 * inv_r_z - psi_g))

        return scalar_gaussian


class AstigmaticGaussianBeamProfile(BeamProfile):
    """Component for constructing astigmatic Gaussian beam data. The normal direction is implicitly
    defined by the ``size`` parameter.

    See also :class:`.AstigmaticGaussianBeam`.
    """

    waist_sizes: tuple[pd.PositiveFloat, pd.PositiveFloat] = pd.Field(
        (1.0, 1.0),
        title="Waist sizes",
        description="Size of the beam at the waist in the local x and y directions.",
        units=MICROMETER,
    )

    waist_distances: tuple[float, float] = pd.Field(
        (0.0, 0.0),
        title="Waist distances",
        description="Distance to the beam waist along the propagation direction "
        "for the waist sizes in the local x and y directions. "
        "When ``direction`` is ``+`` and ``waist_distances`` are positive, the waist "
        "is on the ``-`` side (behind) the beam plane. When ``direction`` is ``+`` and "
        "``waist_distances`` are negative, the waist is on the ``+`` side (in front) of "
        "the beam plane.",
        units=MICROMETER,
    )

    def beam_params(self, z: Numpy, k0: Numpy) -> tuple[Numpy, Numpy, Numpy, Numpy]:
        """Compute the parameters needed to evaluate an astigmatic Gaussian beam at z.

        Parameters
        ----------
        z : Numpy
            Axial distance from the beam center.
        k0 : Numpy
            Wave vector magnitude.
        """

        w_xy, z_xy = self.waist_sizes, self.waist_distances  # shape (2, )
        z_r = [w**2 * k0 / 2 for w in w_xy]  # shape (2, Nk0)
        w_z, w_0, inv_r_z, psi_g = [], [], [], []  # final shape (2, Np, Nk0) after loop below
        for w, z_i, z_ri in zip(w_xy, z_xy, z_r):
            w_z.append(w * np.sqrt(1 + ((z + z_i) / z_ri) ** 2))
            w_0.append(w * np.sqrt(1 + (z_i / z_ri) ** 2))
            inv_r_z.append((z + z_i) / ((z + z_i) ** 2 + z_ri**2))
            # we choose gauge such that psi_g = 0 at z = 0 (beam plane)
            # this is needed for a proper interpolation between different frequencies
            # psi_g shape (Np, Nk0)
            psi_g_1 = np.arctan((z + z_i) / z_ri) / 2
            psi_g_2 = np.arctan(z_i / z_ri) / 2
            psi_g.append(psi_g_1 - psi_g_2)

        return w_0, w_z, inv_r_z, psi_g

    def scalar_field(self, points: Numpy, background_n: float) -> Numpy:
        """
        Scalar field for astigmatic Gaussian beam.
        Scalar field corresponding to the analytic beam in coordinate system such that the
        propagation direction is z and the ``E``-field is entirely ``x``-polarized. The field is
        computed on an unstructured array ``points`` of shape ``(3, ...)``.
        """
        k0 = 2 * np.pi * np.array(self.freqs) / C_0 * background_n
        x, y, z = points
        w_0, w_z, inv_r_z, psi_g = self.beam_params(z, k0)
        x_2 = x**2
        y_2 = y**2

        q_term1_inv = 1j * x_2 * k0 / 2 * inv_r_z[0] - x_2 / w_z[0] ** 2
        q_term2_inv = 1j * y_2 * k0 / 2 * inv_r_z[1] - y_2 / w_z[1] ** 2
        angle_term = q_term1_inv + q_term2_inv + 1j * (z * k0 - psi_g[0] - psi_g[1])
        scalar_gaussian = np.exp(angle_term)

        ampl = np.sqrt(w_0[0] * w_0[1] / w_z[0] / w_z[1])
        return scalar_gaussian * ampl



from __future__ import annotations

from html import escape

from tidy3d.exceptions import SetupError


def plot_scene_3d(scene, width=800, height=800) -> None:
    import gzip
    import json
    from base64 import b64encode
    from io import BytesIO

    import h5py

    # Serialize scene to HDF5 in-memory
    buffer = BytesIO()
    scene.to_hdf5(buffer)
    buffer.seek(0)

    # Open source HDF5 for reading and prepare modified copy
    with h5py.File(buffer, "r") as src:
        buffer2 = BytesIO()
        with h5py.File(buffer2, "w") as dst:

            def copy_item(name, obj):
                if isinstance(obj, h5py.Group):
                    dst.create_group(name)
                    for k, v in obj.attrs.items():
                        dst[name].attrs[k] = v
                elif isinstance(obj, h5py.Dataset):
                    data = obj[()]
                    if name == "JSON_STRING":
                        # Parse and update JSON string
                        json_str = (
                            data.decode("utf-8") if isinstance(data, (bytes, bytearray)) else data
                        )
                        json_data = json.loads(json_str)
                        json_data["size"] = list(scene.size)
                        json_data["center"] = list(scene.center)
                        json_data["grid_spec"] = {}
                        new_str = json.dumps(json_data)
                        dst.create_dataset(name, data=new_str.encode("utf-8"))
                    else:
                        dst.create_dataset(name, data=data)
                        for k, v in obj.attrs.items():
                            dst[name].attrs[k] = v

            src.visititems(copy_item)
        buffer2.seek(0)

    # Gzip the modified HDF5
    gz_buffer = BytesIO()
    with gzip.GzipFile(fileobj=gz_buffer, mode="wb") as gz:
        gz.write(buffer2.read())
    gz_buffer.seek(0)

    # Base64 encode and display with gzipped flag
    sim_base64 = b64encode(gz_buffer.read()).decode("utf-8")
    plot_sim_3d(sim_base64, width=width, height=height, is_gz_base64=True)


def plot_sim_3d(sim, width=800, height=800, is_gz_base64=False) -> None:
    """Make 3D display of simulation in ipython notebook."""

    try:
        from IPython.display import HTML, display
    except ImportError as e:
        raise SetupError(
            "3D plotting requires ipython to be installed "
            "and the code to be running on a jupyter notebook."
        ) from e

    from base64 import b64encode
    from io import BytesIO

    if not is_gz_base64:
        buffer = BytesIO()
        sim.to_hdf5_gz(buffer)
        buffer.seek(0)
        base64 = b64encode(buffer.read()).decode("utf-8")
    else:
        base64 = sim

    js_code = """
        /**
        * Simulation Viewer Injector
        *
        * Monitors the document for elements being added in the form:
        *
        *    <div class="simulation-viewer" data-width="800" data-height="800" data-simulation="{...}" />
        *
        * This script will then inject an iframe to the viewer application, and pass it the simulation data
        * via the postMessage API on request. The script may be safely included multiple times, with only the
        * configuration of the first started script (e.g. viewer URL) applying.
        *
        */
        (function() {
            const TARGET_CLASS = "simulation-viewer";
            const ACTIVE_CLASS = "simulation-viewer-active";
            const VIEWER_URL = "https://tidy3d.simulation.cloud/simulation-viewer";

            class SimulationViewerInjector {
                constructor() {
                    for (var node of document.getElementsByClassName(TARGET_CLASS)) {
                        this.injectViewer(node);
                    }

                    // Monitor for newly added nodes to the DOM
                    this.observer = new MutationObserver(this.onMutations.bind(this));
                    this.observer.observe(document.body, {childList: true, subtree: true});
                }

                onMutations(mutations) {
                    for (var mutation of mutations) {
                        if (mutation.type === 'childList') {
                            /**
                            * Have found that adding the element does not reliably trigger the mutation observer.
                            * It may be the case that setting content with innerHTML does not trigger.
                            *
                            * It seems to be sufficient to re-scan the document for un-activated viewers
                            * whenever an event occurs, as Jupyter triggers multiple events on cell evaluation.
                            */
                            var viewers = document.getElementsByClassName(TARGET_CLASS);
                            for (var node of viewers) {
                                this.injectViewer(node);
                            }
                        }
                    }
                }

                injectViewer(node) {
                    // (re-)check that this is a valid simulation container and has not already been injected
                    if (node.classList.contains(TARGET_CLASS) && !node.classList.contains(ACTIVE_CLASS)) {
                        // Mark node as injected, to prevent re-runs
                        node.classList.add(ACTIVE_CLASS);

                        var uuid;
                        if (window.crypto && window.crypto.randomUUID) {
                            uuid = window.crypto.randomUUID();
                        } else {
                            uuid = "" + Math.random();
                        }

                        var frame = document.createElement("iframe");
                        frame.width = node.dataset.width || 800;
                        frame.height = node.dataset.height || 800;
                        frame.style.cssText = `width:${frame.width}px;height:${frame.height}px;max-width:none;border:0;display:block`
                        frame.src = VIEWER_URL + "?uuid=" + uuid;

                        var postMessageToViewer;
                        postMessageToViewer = event => {
                            if(event.data.type === 'viewer' && event.data.uuid===uuid){
                                frame.contentWindow.postMessage({ type: 'jupyter', uuid, value: node.dataset.simulation, fileType: 'hdf5'}, '*');

                                // Run once only
                                window.removeEventListener('message', postMessageToViewer);
                            }
                        };
                        window.addEventListener(
                            'message',
                            postMessageToViewer,
                            false
                        );

                        node.appendChild(frame);
                    }
                }
            }

            if (!window.simulationViewerInjector) {
                window.simulationViewerInjector = new SimulationViewerInjector();
            }
        })();
    """
    html_code = f"""
    <div class="simulation-viewer" data-width="{escape(str(width))}" data-height="{escape(str(height))}" data-simulation="{escape(base64)}" ></div>
    <script>
        {js_code}
    </script>
    """

    return display(HTML(html_code))



"""Exact and approximate formulas in microwave engineering.

References
----------
[1]     F. E. Terman, Radio engineers' handbook, McGraw-Hill Book Company, Inc., 1943.

[2]     E. B. Rosa, The self and mutal inductances of linear conductors, US Department of
        Commerce and Labor, Bureau of Standards, 1908.

[3]     Y. Y. Iossel, E. S. Kochanov, and M. G. Strunskii, “The Calculation of Electrical Capacitance,”
        Foreign Technology Division Air Force Systems Command U.S. Air Force, 1971.
"""

from __future__ import annotations

import numpy as np

from tidy3d.components.geometry.base import Geometry
from tidy3d.components.types import Axis
from tidy3d.constants import EPSILON_0


def inductance_straight_rectangular_wire(
    size: tuple[float, float, float], current_axis: Axis
) -> float:
    """Computes the self-inductance of a finite length of wire with a rectangular cross section.
    Equation 26 [1] and Equation 21 [2].

    Parameters
    ----------
    size :
        Tuple representing the size of the rectangular wire segment in microns.
    current_axis :
        The axis along which the current is flowing.

    Returns
    -------
    float
        The self-inductance in Henrys."""

    length, transverse_sizes = Geometry.pop_axis(size, current_axis)
    b_plus_c = transverse_sizes[0] + transverse_sizes[1]
    log_term = np.log(2 * length / b_plus_c) + 0.5 + 0.2235 * (b_plus_c / length)
    L0 = 2 * (length) * log_term * 1e-13
    return L0


def mutual_inductance_colinear_wire_segments(l1: float, l2: float, d: float) -> float:
    """Computes the mutual inductance between two wire segments arranged in the same line,
    Equation 73 [1].

    Parameters
    ----------
    l1 :
        Length of the first segment in microns.
    l2 :
        Length of the second segment in microns.
    d :
        Separation distance in microns between segment end points.

    Returns
    -------
    float
        The mutual inductance in Henrys."""

    sum_all = l1 + l2 + d
    sum_l1 = l1 + d
    sum_l2 = l2 + d
    M = (
        1e-7
        * (
            sum_all * np.log(sum_all)
            + d * np.log(d)
            - sum_l1 * np.log(sum_l1)
            - sum_l2 * np.log(sum_l2)
        )
        * 1e-6
    )
    return M


def total_inductance_colinear_rectangular_wire_segments(
    first: tuple[float, float, float],
    second: tuple[float, float, float],
    d: float,
    current_axis: Axis,
) -> float:
    """Computes the total inductance of a pair of finite length of colinear wire segments with a
    rectangular cross section, Equation 69 [1].

    Parameters
    ----------
    first :
        Size of the first rectangular wire segment.
    second :
        Size of the the second rectangular wire segment.
    d :
        Separation distance in microns between wire end points.
    current_axis :
        The axis along which the current is flowing.

    Returns
    -------
    float
        The total inductance in Henrys."""

    Lfirst = inductance_straight_rectangular_wire(first, current_axis)
    Lsecond = inductance_straight_rectangular_wire(second, current_axis)
    M = mutual_inductance_colinear_wire_segments(first[current_axis], second[current_axis], d)
    return Lfirst + Lsecond + 2 * M


def capacitance_rectangular_sheets(
    width: float, length: float, d: float, eps_r: float = 1
) -> float:
    """Computes the capacitance between a pair of rectangular sheets, Equation 4-28 [3].

    Parameters
    ----------
    width :
        Width of sheets in microns.
    length :
        Length of each sheet in microns.
    d :
        Separation distance in microns between wire end points.

    Returns
    -------
    float
        The capacitance in Farads."""
    # Approximation valid when sheets are thin, width >> length, and length >> d
    C = 2 / np.pi * EPSILON_0 * eps_r * width * np.log(4 * (1 + 2 * length / d))
    return C


def capacitance_colinear_cylindrical_wire_segments(
    radius: float, length: float, d: float, eps_r: float = 1
) -> float:
    """Computes the capacitance between a pair of colinear cylindrical wires of finite length,
    Equation 3-43 [3].

    Parameters
    ----------
    radius :
        Radius of the wire segment in microns.
    length :
        Length of each wire segment in microns.
    d :
        Separation distance in microns between wire end points.

    Returns
    -------
    float
        The capacitance in Farads."""
    # Approximation valid when wires are thin and long (radius << length)

    m = d / 2
    if m < length:
        ratio = m / length
        D2 = (
            0.434
            + ratio * np.log10(4 * ratio)
            + (1 + ratio) * np.log10(1 + ratio)
            - (1 + 2 * ratio) * np.log10(1 + 2 * ratio)
        )
    else:
        ml = m / length
        lm = length / m
        D2 = (
            0.133
            + ml * (1 + lm) * np.log10(1 + lm)
            - (2 * ml) * (1 + lm / 2) * np.log10(1 + lm / 2)
        )
    C = np.pi * EPSILON_0 * eps_r * length / (np.log(length / radius) - 2.303 * D2)
    return C



"""Class and custom data array for representing a scattering matrix port based on lumped circuit elements."""

from __future__ import annotations

from abc import abstractmethod
from typing import Optional

import pydantic.v1 as pd

from tidy3d.components.base import cached_property
from tidy3d.components.geometry.utils_2d import snap_coordinate_to_grid
from tidy3d.components.grid.grid import Grid, YeeGrid
from tidy3d.components.lumped_element import LumpedElementType
from tidy3d.components.monitor import FieldMonitor
from tidy3d.components.types import Complex, Coordinate, FreqArray
from tidy3d.constants import OHM

from .base_terminal import AbstractTerminalPort

DEFAULT_PORT_NUM_CELLS = 3
DEFAULT_REFERENCE_IMPEDANCE = 50


class AbstractLumpedPort(AbstractTerminalPort):
    """Class representing a single lumped port."""

    impedance: Complex = pd.Field(
        DEFAULT_REFERENCE_IMPEDANCE,
        title="Reference impedance",
        description="Reference port impedance for scattering parameter computation.",
        units=OHM,
    )

    num_grid_cells: Optional[pd.PositiveInt] = pd.Field(
        DEFAULT_PORT_NUM_CELLS,
        title="Port grid cells",
        description="Number of mesh grid cells associated with the port along each direction, "
        "which are added through automatic mesh refinement. "
        "A value of ``None`` will turn off automatic mesh refinement.",
    )

    enable_snapping_points: bool = pd.Field(
        True,
        title="Snap Grid To Lumped Port",
        description="When enabled, snapping points are automatically generated to snap grids to key "
        "geometric features of the lumped port for more accurate modelling.",
    )

    @cached_property
    def _voltage_monitor_name(self) -> str:
        return f"{self.name}_voltage"

    @cached_property
    def _current_monitor_name(self) -> str:
        return f"{self.name}_current"

    def snapped_center(self, grid: Grid) -> Coordinate:
        """Get the exact center of this port after snapping along the injection axis.
        Ports are snapped to the nearest Yee cell boundary to match the exact position
        of the load.
        """
        center = list(self.center)
        normal_axis = self.injection_axis
        normal_port_center = center[normal_axis]
        center[normal_axis] = snap_coordinate_to_grid(grid, normal_port_center, normal_axis)
        return tuple(center)

    @cached_property
    @abstractmethod
    def to_load(self, snap_center: Optional[float] = None) -> LumpedElementType:
        """Create a load from the lumped port."""

    @abstractmethod
    def to_voltage_monitor(
        self, freqs: FreqArray, snap_center: Optional[float] = None
    ) -> FieldMonitor:
        """Field monitor to compute port voltage."""

    @abstractmethod
    def to_current_monitor(
        self, freqs: FreqArray, snap_center: Optional[float] = None
    ) -> FieldMonitor:
        """Field monitor to compute port current."""

    def to_monitors(
        self, freqs: FreqArray, snap_center: Optional[float] = None, grid: Grid = None
    ) -> list[FieldMonitor]:
        """Field monitors to compute port voltage and current."""
        return [
            self.to_voltage_monitor(freqs, snap_center, grid),
            self.to_current_monitor(freqs, snap_center, grid),
        ]

    @abstractmethod
    def _check_grid_size(self, yee_grid: YeeGrid):
        """Raises :class:`SetupError` if the grid is too coarse at port locations."""



"""
This module provides integration between Tidy3D and KLayout.
"""

from __future__ import annotations

from .drc import DRCConfig, DRCResults, DRCRunner, run_drc_on_gds
from .util import check_installation

__all__ = [
    "DRCConfig",
    "DRCResults",
    "DRCRunner",
    "check_installation",
    "run_drc_on_gds",
]



"""Holds the reference materials for Tidy3D material library."""

from __future__ import annotations

from tidy3d.material_library.material_reference import ReferenceData

rf_material_refs = {
    "Rogers3003": ReferenceData(
        manufacturer="Rogers Corporation",
        datasheet_title="RO3003™ Laminates",
        url="https://www.rogerscorp.com/advanced-electronics-solutions/ro3000-series-laminates/ro3003-laminates",
    ),
    "Rogers3010": ReferenceData(
        manufacturer="Rogers Corporation",
        datasheet_title="RO3010™ Laminates",
        url="https://www.rogerscorp.com/advanced-electronics-solutions/ro3000-series-laminates/ro3010-laminates",
    ),
    "Rogers4003C": ReferenceData(
        manufacturer="Rogers Corporation",
        datasheet_title="RO4003C™ Laminates",
        url="https://www.rogerscorp.com/advanced-electronics-solutions/ro4000-series-laminates/ro4350b-laminates",
    ),
    "Rogers4350B": ReferenceData(
        manufacturer="Rogers Corporation",
        datasheet_title="RO4350B™ Laminates",
        url="https://www.rogerscorp.com/advanced-electronics-solutions/ro4000-series-laminates/ro4350b-laminates",
    ),
    "ArlonAD255C": ReferenceData(
        manufacturer="Rogers Corporation",
        datasheet_title="AD255C High Performance Polyimide Laminates",
        url="https://www.rogerscorp.com/advanced-electronics-solutions/ad-series-laminates/ad255c-laminates",
    ),
    "FR4_standard": ReferenceData(
        manufacturer="Isola",
        datasheet_title="Standard FR-4 Epoxy Glass Cloth Laminate",
        url="https://www.isola-group.com/pcb-laminates-prepreg/is410-fr-4-epoxy-laminate-and-prepreg/",
    ),
    "FR4_lowloss": ReferenceData(
        manufacturer="Isola",
        datasheet_title="Low loss FR-4 Epoxy Glass Cloth Laminate",
        url="https://www.isola-group.com/pcb-laminates-prepreg/is410-fr-4-epoxy-laminate-and-prepreg/",
    ),
}



"""Defines heat simulation class
NOTE: Keeping this class for backward compatibility only"""

from __future__ import annotations

from typing import Optional

import pydantic.v1 as pd

from tidy3d.components.tcad.simulation.heat_charge import HeatChargeSimulation
from tidy3d.components.types import Ax
from tidy3d.components.viz import add_ax_if_none, equal_aspect
from tidy3d.log import log


class HeatSimulation(HeatChargeSimulation):
    """
    Contains all information about heat simulation.

    Example
    -------
    >>> import tidy3d as td
    >>> heat_sim = td.HeatSimulation( # doctest: +SKIP
    ...     size=(3.0, 3.0, 3.0),
    ...     structures=[
    ...         td.Structure(
    ...             geometry=td.Box(size=(1, 1, 1), center=(0, 0, 0)),
    ...             medium=td.Medium(
    ...                 permittivity=2.0, heat_spec=td.SolidSpec(
    ...                     conductivity=1,
    ...                     capacity=1,
    ...                 )
    ...             ),
    ...             name="box",
    ...         ),
    ...     ],
    ...     medium=td.Medium(permittivity=3.0, heat_spec=td.FluidSpec()),
    ...     grid_spec=td.UniformUnstructuredGrid(dl=0.1),
    ...     sources=[td.HeatSource(rate=1, structures=["box"])],
    ...     boundary_spec=[
    ...         td.HeatChargeBoundarySpec(
    ...             placement=td.StructureBoundary(structure="box"),
    ...             condition=td.TemperatureBC(temperature=500),
    ...         )
    ...     ],
    ...     monitors=[td.TemperatureMonitor(size=(1, 2, 3), name="sample")],
    ... )
    """

    @pd.root_validator(skip_on_failure=True)
    def issue_warning_deprecated(cls, values):
        """Issue warning for 'HeatSimulations'."""
        log.warning(
            "Setting up deprecated 'HeatSimulation'. "
            "Consider defining 'HeatChargeSimulation' instead."
        )
        return values

    @equal_aspect
    @add_ax_if_none
    def plot_heat_conductivity(
        self,
        x: Optional[float] = None,
        y: Optional[float] = None,
        z: Optional[float] = None,
        ax: Ax = None,
        alpha: Optional[float] = None,
        source_alpha: Optional[float] = None,
        monitor_alpha: Optional[float] = None,
        colorbar: str = "conductivity",
        hlim: Optional[tuple[float, float]] = None,
        vlim: Optional[tuple[float, float]] = None,
    ) -> Ax:
        """Plot each of simulation's components on a plane defined by one nonzero x,y,z coordinate.

        Parameters
        ----------
        x : float = None
            position of plane in x direction, only one of x, y, z must be specified to define plane.
        y : float = None
            position of plane in y direction, only one of x, y, z must be specified to define plane.
        z : float = None
            position of plane in z direction, only one of x, y, z must be specified to define plane.
        ax : matplotlib.axes._subplots.Axes = None
            Matplotlib axes to plot on, if not specified, one is created.
        alpha : float = None
            Opacity of the structures being plotted.
            Defaults to the structure default alpha.
        source_alpha : float = None
            Opacity of the sources. If ``None``, uses Tidy3d default.
        monitor_alpha : float = None
            Opacity of the monitors. If ``None``, uses Tidy3d default.
        colorbar: str = "conductivity"
            Display colorbar for thermal conductivity ("conductivity") or heat source rate
            ("source").
        hlim : Tuple[float, float] = None
            The x range if plotting on xy or xz planes, y range if plotting on yz plane.
        vlim : Tuple[float, float] = None
            The z range if plotting on xz or yz planes, y plane if plotting on xy plane.

        Returns
        -------
        matplotlib.axes._subplots.Axes
            The supplied or created matplotlib axes.
        """

        plot_type = None
        if colorbar == "conductivity":
            plot_type = "heat_conductivity"
        elif colorbar == "source":
            plot_type = "source"

        return self.plot_property(
            x=x,
            y=y,
            z=z,
            ax=ax,
            alpha=alpha,
            source_alpha=source_alpha,
            monitor_alpha=monitor_alpha,
            property=plot_type,
            hlim=hlim,
            vlim=vlim,
        )



from __future__ import annotations

from typing import Union

from tidy3d.components.spice.analysis.dc import IsothermalSteadyChargeDCAnalysis

ElectricalAnalysisType = Union[IsothermalSteadyChargeDCAnalysis]



"""Imports from scattering matrix plugin."""

from __future__ import annotations

import warnings

from tidy3d.plugins.smatrix.component_modelers.base import (
    AbstractComponentModeler,
)
from tidy3d.plugins.smatrix.component_modelers.modal import ModalComponentModeler
from tidy3d.plugins.smatrix.component_modelers.terminal import TerminalComponentModeler
from tidy3d.plugins.smatrix.component_modelers.types import ComponentModelerType
from tidy3d.plugins.smatrix.data.data_array import (
    ModalPortDataArray,
    PortDataArray,
    TerminalPortDataArray,
)
from tidy3d.plugins.smatrix.data.modal import ModalComponentModelerData
from tidy3d.plugins.smatrix.data.terminal import MicrowaveSMatrixData, TerminalComponentModelerData
from tidy3d.plugins.smatrix.data.types import ComponentModelerDataType
from tidy3d.plugins.smatrix.ports.coaxial_lumped import CoaxialLumpedPort
from tidy3d.plugins.smatrix.ports.modal import Port
from tidy3d.plugins.smatrix.ports.rectangular_lumped import LumpedPort
from tidy3d.plugins.smatrix.ports.wave import WavePort

# Instantiate on plugin import till we unite with toplevel
warnings.filterwarnings(
    "once",
    message="ℹ️ ⚠️ RF simulations are subject to new license requirements in the future. You have instantiated at least one RF-specific component.",
    category=FutureWarning,
)

# Legacy type to support previous flows
ComponentModeler = ModalComponentModeler

__all__ = [
    "AbstractComponentModeler",
    "CoaxialLumpedPort",
    "ComponentModeler",
    "ComponentModelerDataType",
    "ComponentModelerType",
    "LumpedPort",
    "MicrowaveSMatrixData",
    "ModalComponentModeler",
    "ModalComponentModelerData",
    "ModalPortDataArray",
    "Port",
    "PortDataArray",
    "TerminalComponentModeler",
    "TerminalComponentModelerData",
    "TerminalPortDataArray",
    "WavePort",
]



"""Utilities for type & schema creation."""

from __future__ import annotations


def _add_schema(arbitrary_type: type, title: str, field_type_str: str) -> None:
    """Adds a schema to the ``arbitrary_type`` class without subclassing."""

    @classmethod
    def mod_schema_fn(cls, field_schema: dict) -> None:
        """Function that gets set to ``arbitrary_type.__modify_schema__``."""
        field_schema.update({"title": title, "type": field_type_str})

    arbitrary_type.__modify_schema__ = mod_schema_fn



"""Objects that define how data is recorded from simulation."""

from __future__ import annotations

from typing import Literal

import pydantic.v1 as pd

from tidy3d.components.tcad.monitors.abstract import HeatChargeMonitor


class SteadyPotentialMonitor(HeatChargeMonitor):
    """
    Electric potential (:math:`\\psi`) monitor.

    Example
    -------
    >>> import tidy3d as td
    >>> voltage_monitor_z0 = td.SteadyPotentialMonitor(
    ... center=(0, 0.14, 0), size=(0.6, 0.3, 0), name="voltage_z0", unstructured=True,
    ... )
    """


class SteadyFreeCarrierMonitor(HeatChargeMonitor):
    """
    Free-carrier monitor for Charge simulations.

    Example
    -------
    >>> import tidy3d as td
    >>> voltage_monitor_z0 = td.SteadyFreeCarrierMonitor(
    ... center=(0, 0.14, 0), size=(0.6, 0.3, 0), name="voltage_z0", unstructured=True,
    ... )
    """

    # NOTE: for the time being supporting unstructured
    unstructured: Literal[True] = pd.Field(
        True,
        title="Unstructured Grid",
        description="Return data on the original unstructured grid.",
    )


class SteadyEnergyBandMonitor(HeatChargeMonitor):
    """
    Energy bands monitor for Charge simulations.

    Example
    -------
    >>> import tidy3d as td
    >>> energy_monitor_z0 = td.SteadyEnergyBandMonitor(
    ... center=(0, 0.14, 0), size=(0.6, 0.3, 0), name="bands_z0", unstructured=True,
    ... )
    """

    # NOTE: for the time being supporting unstructured
    unstructured: Literal[True] = pd.Field(
        True,
        title="Unstructured Grid",
        description="Return data on the original unstructured grid.",
    )


class SteadyCapacitanceMonitor(HeatChargeMonitor):
    """
    Capacitance monitor associated with a charge simulation.

    Example
    -------
    >>> import tidy3d as td
    >>> capacitance_global_mnt = td.SteadyCapacitanceMonitor(
    ... center=(0, 0.14, 0), size=(td.inf, td.inf, 0), name="capacitance_global_mnt",
    ... )
    """

    # NOTE: for the time being supporting unstructured
    unstructured: Literal[True] = pd.Field(
        True,
        title="Unstructured Grid",
        description="Return data on the original unstructured grid.",
    )


class SteadyElectricFieldMonitor(HeatChargeMonitor):
    """
    Electric field monitor for Charge/Conduction simulations.

    Example
    -------
    >>> import tidy3d as td
    >>> electric_field_monitor_z0 = td.SteadyElectricFieldMonitor(
    ... center=(0, 0.14, 0), size=(0.6, 0.3, 0), name="electric_field_z0",
    ... )
    """

    unstructured: Literal[True] = pd.Field(
        True,
        title="Unstructured Grid",
        description="Return data on the original unstructured grid.",
    )


class SteadyCurrentDensityMonitor(HeatChargeMonitor):
    """
    Current density monitor for Charge/Conduction simulations.

    Example
    -------
    >>> import tidy3d as td
    >>> current_density_monitor_z0 = td.SteadyCurrentDensityMonitor(
    ... center=(0, 0.14, 0), size=(0.6, 0.3, 0), name="current_density_z0",
    ... )
    """

    unstructured: Literal[True] = pd.Field(
        True,
        title="Unstructured Grid",
        description="Return data on the original unstructured grid.",
    )



"""Defines a jax-compatible simulation."""

from __future__ import annotations

from typing import Literal, Optional, Union

import numpy as np
import pydantic.v1 as pd
import xarray as xr
from jax.tree_util import register_pytree_node_class
from joblib import Parallel, delayed

from tidy3d.components.base import Tidy3dBaseModel, cached_property, skip_if_fields_missing
from tidy3d.components.data.monitor_data import FieldData, PermittivityData
from tidy3d.components.geometry.base import Box
from tidy3d.components.medium import AbstractMedium
from tidy3d.components.monitor import (
    DiffractionMonitor,
    FieldMonitor,
    ModeMonitor,
    Monitor,
    PermittivityMonitor,
)
from tidy3d.components.simulation import Simulation
from tidy3d.components.structure import Structure
from tidy3d.components.subpixel_spec import Staircasing, SubpixelSpec
from tidy3d.components.types import Ax, annotate_type
from tidy3d.constants import HERTZ, SECOND
from tidy3d.exceptions import AdjointError
from tidy3d.log import log

from .base import WEB_ADJOINT_MESSAGE, JaxObject
from .geometry import JaxGeometryGroup, JaxPolySlab
from .structure import (
    JaxStructure,
    JaxStructureStaticGeometry,
    JaxStructureStaticMedium,
    JaxStructureType,
)

# bandwidth of adjoint source in units of freq0 if no `fwidth_adjoint`, and one output freq
FWIDTH_FACTOR = 1.0 / 10

# bandwidth of adjoint sources in units of the minimum difference between output frequencies
FWIDTH_FACTOR_MULTIFREQ = 0.1

# the adjoint run time is the forward simulation run time + RUN_TIME_FACTOR / fwidth
RUN_TIME_FACTOR = 10

# how many processors to use for server and client side adjoint
NUM_PROC_LOCAL = 1

# number of input structures before it errors
MAX_NUM_INPUT_STRUCTURES = 400

# generic warning for nonlinearity
NL_WARNING = (
    "The 'adjoint' plugin does not currently support nonlinear materials. "
    "While the gradients might be calculated, they will be inaccurate and the "
    "error will increase as the strength of the nonlinearity is increased. "
    "We strongly recommend using linear simulations only with the adjoint plugin."
)

OutputMonitorTypes = (DiffractionMonitor, FieldMonitor, ModeMonitor)
OutputMonitorType = tuple[annotate_type(Union[OutputMonitorTypes]), ...]


class JaxInfo(Tidy3dBaseModel):
    """Class to store information when converting between jax and tidy3d."""

    num_input_structures: pd.NonNegativeInt = pd.Field(
        ...,
        title="Number of Input Structures",
        description="Number of input structures in the original JaxSimulation.",
    )

    num_output_monitors: pd.NonNegativeInt = pd.Field(
        ...,
        title="Number of Output Monitors",
        description="Number of output monitors in the original JaxSimulation.",
    )

    num_grad_monitors: pd.NonNegativeInt = pd.Field(
        ...,
        title="Number of Gradient Monitors",
        description="Number of gradient monitors in the original JaxSimulation.",
    )

    num_grad_eps_monitors: pd.NonNegativeInt = pd.Field(
        ...,
        title="Number of Permittivity Monitors",
        description="Number of permittivity monitors in the original JaxSimulation.",
    )

    fwidth_adjoint: float = pd.Field(
        None,
        title="Adjoint Frequency Width",
        description="Custom frequency width of the original JaxSimulation.",
        units=HERTZ,
    )

    run_time_adjoint: float = pd.Field(
        None,
        title="Adjoint Run Time",
        description="Custom run time of the original JaxSimulation.",
        units=SECOND,
    )

    input_structure_types: tuple[
        Literal["JaxStructure", "JaxStructureStaticMedium", "JaxStructureStaticGeometry"], ...
    ] = pd.Field(
        (),
        title="Input Structure Types",
        description="Type of the original input_structures (as strings).",
    )


@register_pytree_node_class
class JaxSimulation(Simulation, JaxObject):
    """A :class:`.Simulation` registered with jax."""

    input_structures: tuple[annotate_type(JaxStructureType), ...] = pd.Field(
        (),
        title="Input Structures",
        description="Tuple of jax-compatible structures"
        " that may depend on differentiable parameters.",
        jax_field=True,
    )

    output_monitors: OutputMonitorType = pd.Field(
        (),
        title="Output Monitors",
        description="Tuple of monitors whose data the differentiable output depends on.",
    )

    grad_monitors: tuple[FieldMonitor, ...] = pd.Field(
        (),
        title="Gradient Field Monitors",
        description="Tuple of monitors used for storing fields, used internally for gradients.",
    )

    grad_eps_monitors: tuple[PermittivityMonitor, ...] = pd.Field(
        (),
        title="Gradient Permittivity Monitors",
        description="Tuple of monitors used for storing epsilon, used internally for gradients.",
    )

    fwidth_adjoint: pd.PositiveFloat = pd.Field(
        None,
        title="Adjoint Frequency Width",
        description="Custom frequency width to use for ``source_time`` of adjoint sources. "
        "If not supplied or ``None``, uses the average fwidth of the original simulation's sources.",
        units=HERTZ,
    )

    run_time_adjoint: pd.PositiveFloat = pd.Field(
        None,
        title="Adjoint Run Time",
        description="Custom ``run_time`` to use for adjoint simulation. "
        "If not supplied or ``None``, uses a factor times the adjoint source ``fwidth``.",
        units=SECOND,
    )

    @pd.validator("output_monitors", always=True)
    def _output_monitors_colocate_false(cls, val):
        """Make sure server-side colocation is off."""
        new_vals = []
        for mnt in val:
            if mnt.colocate:
                log.warning(
                    "Output field monitors in the adjoint plugin require explicitly setting "
                    "'colocate=False'. Setting 'colocate=False' in monitor '{mnt.name}'. ",
                    "Use 'SimulationData.at_boundaries' after the solver run to automatically "
                    "colocate the fields to the grid boundaries, or 'MonitorData.colocate' "
                    "if colocating to custom coordinates.",
                )
                mnt = mnt.updated_copy(colocate=False)
            new_vals.append(mnt)
        return new_vals

    @pd.validator("subpixel", always=True)
    def _subpixel_is_on(cls, val):
        """Assert dielectric subpixel is on."""
        if (isinstance(val, SubpixelSpec) and isinstance(val.dielectric, Staircasing)) or not val:
            raise AdjointError(
                "'JaxSimulation.subpixel' must be 'True' or a specific 'SubpixelSpec' "
                "with no dielectric staircasing to use adjoint plugin."
            )
        return val

    @pd.validator("input_structures", always=True)
    @skip_if_fields_missing(["structures"])
    def _warn_overlap(cls, val, values):
        """Print appropriate warning if structures intersect in ways that cause gradient error."""

        input_structures = [s for s in val if "geometry" in s._differentiable_fields]

        structures = list(values.get("structures"))

        # if the center and size of all structure geometries do not contain all numbers, skip check
        for struct in input_structures:
            geometry = struct.geometry
            size_all_floats = all(isinstance(s, (float, int)) for s in geometry.bound_size)
            cent_all_floats = all(isinstance(c, (float, int)) for c in geometry.bound_center)
            if not (size_all_floats and cent_all_floats):
                return val

        with log as consolidated_logger:
            # check intersections with other input_structures
            for i, in_struct_i in enumerate(input_structures):
                geometry_i = in_struct_i.geometry
                for j in range(i + 1, len(input_structures)):
                    if geometry_i.intersects(input_structures[j].geometry):
                        consolidated_logger.warning(
                            f"'JaxSimulation.input_structures[{i}]' overlaps or touches "
                            f"'JaxSimulation.input_structures[{j}]'. Geometric gradients for "
                            "overlapping input structures may contain errors.",
                            log_once=True,
                        )

            # check JaxPolySlab intersections with background structures
            for i, in_struct_i in enumerate(input_structures):
                geometry_i = in_struct_i.geometry
                if not isinstance(geometry_i, JaxPolySlab):
                    continue
                for j, struct_j in enumerate(structures):
                    if geometry_i.intersects(struct_j.geometry):
                        consolidated_logger.warning(
                            f"'JaxPolySlab'-containing 'JaxSimulation.input_structures[{i}]' "
                            f"intersects with 'JaxSimulation.structures[{j}]'. Note that in this "
                            "version of the adjoint plugin, there may be errors in the gradient "
                            "when 'JaxPolySlab' intersects with background structures."
                        )

        return val

    @pd.validator("output_monitors", always=True)
    def _warn_if_colocate(cls, val):
        """warn if any colocate=True in output FieldMonitors."""
        for index, mnt in enumerate(val):
            if isinstance(mnt, FieldMonitor):
                if mnt.colocate:
                    log.warning(
                        f"'FieldMonitor' at 'JaxSimulation.output_monitors[{index}]' "
                        "has 'colocate=True', "
                        "this may lead to decreased accuracy in adjoint gradient."
                    )
                    return val
        return val

    @pd.validator("medium", always=True)
    def _warn_nonlinear_medium(cls, val):
        """warn if the jax simulation medium is nonlinear."""
        # hasattr is just an additional check to avoid unnecessary bugs
        # if a medium is encountered that does not support nonlinear spec, or things change.
        if hasattr(val, "nonlinear_spec") and val.nonlinear_spec:
            log.warning(
                "Nonlinear background medium detected in the 'JaxSimulation'. " + NL_WARNING
            )
        return val

    @pd.validator("structures", always=True)
    def _warn_nonlinear_structure(cls, val):
        """warn if a jax simulation structure.medium is nonlinear."""
        for i, struct in enumerate(val):
            medium = struct.medium
            # hasattr is just an additional check to avoid unnecessary bugs
            # if a medium is encountered that does not support nonlinear spec, or things change.
            if hasattr(medium, "nonlinear_spec") and medium.nonlinear_spec:
                log.warning(f"Nonlinear medium detected in structures[{i}]. " + NL_WARNING)
        return val

    @pd.validator("input_structures", always=True)
    def _warn_nonlinear_input_structure(cls, val):
        """warn if a jax simulation input_structure.medium is nonlinear."""
        for i, struct in enumerate(val):
            medium = struct.medium
            # hasattr is just an additional check to avoid unnecessary bugs
            # if a medium is encountered that does not support nonlinear spec, or things change.
            if hasattr(medium, "nonlinear_spec") and medium.nonlinear_spec:
                log.warning(f"Nonlinear medium detected in input_structures[{i}]. " + NL_WARNING)
        return val

    def _restrict_input_structures(self) -> None:
        """Restrict number of input structures."""
        num_input_structures = len(self.input_structures)
        if num_input_structures > MAX_NUM_INPUT_STRUCTURES:
            raise AdjointError(
                "For performance, adjoint plugin restricts the number of input structures to "
                f"{MAX_NUM_INPUT_STRUCTURES}. Found {num_input_structures}. " + WEB_ADJOINT_MESSAGE
            )

    def _validate_web_adjoint(self) -> None:
        """Run validators for this component, only if using ``tda.web.run()``."""
        self._restrict_input_structures()
        for structure in self.input_structures:
            structure._validate_web_adjoint()

    @staticmethod
    def get_freqs_adjoint(output_monitors: list[Monitor]) -> list[float]:
        """Return sorted list of unique frequencies stripped from a collection of monitors."""

        if len(output_monitors) == 0:
            raise AdjointError("Can't get adjoint frequency as no output monitors present.")

        output_freqs = []
        for mnt in output_monitors:
            for freq in mnt.freqs:
                output_freqs.append(freq)

        return np.unique(output_freqs).tolist()

    @cached_property
    def freqs_adjoint(self) -> list[float]:
        """Return sorted list of frequencies stripped from the output monitors."""
        return self.get_freqs_adjoint(output_monitors=self.output_monitors)

    @cached_property
    def _is_multi_freq(self) -> bool:
        """Does this simulation have a multi-frequency output?"""
        return len(self.freqs_adjoint) > 1

    @cached_property
    def _min_delta_freq(self) -> float:
        """Minimum spacing between output_frequencies (Hz)."""

        if not self._is_multi_freq:
            return None

        delta_freqs = np.abs(np.diff(np.sort(np.array(self.freqs_adjoint))))
        return np.min(delta_freqs)

    @cached_property
    def _fwidth_adjoint(self) -> float:
        """Frequency width to use for adjoint source, user-defined or the average of the sources."""

        # if user-specified, use that
        if self.fwidth_adjoint is not None:
            return self.fwidth_adjoint

        freqs_adjoint = self.freqs_adjoint

        # multiple output frequency case
        if self._is_multi_freq:
            return FWIDTH_FACTOR_MULTIFREQ * self._min_delta_freq

        # otherwise, grab from sources and output monitors
        num_sources = len(self.sources)  # should be 0 for adjoint already but worth checking

        # if no sources, just use a constant factor times the mean adjoint frequency
        if num_sources == 0:
            return FWIDTH_FACTOR * np.mean(freqs_adjoint)

        # if more than one forward source, use their maximum
        if num_sources > 1:
            log.warning(f"{num_sources} sources, using their maximum 'fwidth' for adjoint source.")

        fwidths = [src.source_time.fwidth for src in self.sources]
        return np.max(fwidths)

    @cached_property
    def _run_time_adjoint(self: float) -> float:
        """Return the run time of the adjoint simulation as a function of its fwidth."""

        if self.run_time_adjoint is not None:
            return self.run_time_adjoint

        run_time_fwd = self._run_time
        run_time_adjoint = run_time_fwd + RUN_TIME_FACTOR / self._fwidth_adjoint

        if self._is_multi_freq:
            log.warning(
                f"{len(self.freqs_adjoint)} unique frequencies detected in the output monitors "
                f"with a minimum spacing of {self._min_delta_freq:.3e} (Hz). "
                f"Setting the 'fwidth' of the adjoint sources to {FWIDTH_FACTOR_MULTIFREQ} times "
                f"this value = {self._fwidth_adjoint:.3e} (Hz) to avoid spectral overlap. "
                "To account for this, the corresponding 'run_time' in the adjoint simulation is "
                f"will be set to {run_time_adjoint:3e} "
                f"compared to {self._run_time:3e} in the forward simulation. "
                "If the adjoint 'run_time' is large due to small frequency spacing, "
                "it could be better to instead run one simulation per frequency, "
                "which can be done in parallel using 'tidy3d.plugins.adjoint.web.run_async'."
            )

        return run_time_adjoint

    @cached_property
    def tmesh_adjoint(self) -> np.ndarray:
        """FDTD time stepping points.

        Returns
        -------
        np.ndarray
            Times (seconds) that the simulation time steps through.
        """
        dt = self.dt
        return np.arange(0.0, self._run_time_adjoint + dt, dt)

    @cached_property
    def num_time_steps_adjoint(self) -> int:
        """Number of time steps in the adjoint simulation."""
        return len(self.tmesh_adjoint)

    def to_simulation(self) -> tuple[Simulation, JaxInfo]:
        """Convert :class:`.JaxSimulation` instance to :class:`.Simulation` with an info dict."""

        sim_dict = self.dict(
            exclude={
                "type",
                "structures",
                "monitors",
                "output_monitors",
                "grad_monitors",
                "grad_eps_monitors",
                "input_structures",
                "fwidth_adjoint",
                "run_time_adjoint",
            }
        )
        sim = Simulation.parse_obj(sim_dict)

        # put all structures and monitors in one list
        all_structures = list(self.scene.sorted_structures) + [
            js.to_structure() for js in self.input_structures
        ]
        all_monitors = (
            list(self.monitors)
            + list(self.output_monitors)
            + list(self.grad_monitors)
            + list(self.grad_eps_monitors)
        )

        sim = sim.updated_copy(structures=all_structures, monitors=all_monitors)

        # information about the state of the original JaxSimulation to stash for reconstruction
        jax_info = JaxInfo(
            num_input_structures=len(self.input_structures),
            num_output_monitors=len(self.output_monitors),
            num_grad_monitors=len(self.grad_monitors),
            num_grad_eps_monitors=len(self.grad_eps_monitors),
            fwidth_adjoint=self.fwidth_adjoint,
            run_time_adjoint=self.run_time_adjoint,
            input_structure_types=[s.type for s in self.input_structures],
        )

        return sim, jax_info

    def to_gds(
        self,
        cell,
        x: Optional[float] = None,
        y: Optional[float] = None,
        z: Optional[float] = None,
        permittivity_threshold: pd.NonNegativeFloat = 1,
        frequency: pd.PositiveFloat = 0,
        gds_layer_dtype_map: Optional[
            dict[AbstractMedium, tuple[pd.NonNegativeInt, pd.NonNegativeInt]]
        ] = None,
    ) -> None:
        """Append the simulation structures to a .gds cell.
        Parameters
        ----------
        cell : ``gdstk.Cell``
            Cell object to which the generated polygons are added.
        x : float = None
            Position of plane in x direction, only one of x,y,z can be specified to define plane.
        y : float = None
            Position of plane in y direction, only one of x,y,z can be specified to define plane.
        z : float = None
            Position of plane in z direction, only one of x,y,z can be specified to define plane.
        permittivity_threshold : float = 1
            Permittivity value used to define the shape boundaries for structures with custom
            medim
        frequency : float = 0
            Frequency for permittivity evaluation in case of custom medium (Hz).
        gds_layer_dtype_map : Dict
            Dictionary mapping mediums to GDSII layer and data type tuples.
        """
        sim, _ = self.to_simulation()
        return sim.to_gds(
            cell=cell,
            x=x,
            y=y,
            z=z,
            permittivity_threshold=permittivity_threshold,
            frequency=frequency,
            gds_layer_dtype_map=gds_layer_dtype_map,
        )

    def to_gdstk(
        self,
        x: Optional[float] = None,
        y: Optional[float] = None,
        z: Optional[float] = None,
        permittivity_threshold: pd.NonNegativeFloat = 1,
        frequency: pd.PositiveFloat = 0,
        gds_layer_dtype_map: Optional[
            dict[AbstractMedium, tuple[pd.NonNegativeInt, pd.NonNegativeInt]]
        ] = None,
    ) -> list:
        """Convert a simulation's planar slice to a .gds type polygon list.
        Parameters
        ----------
        x : float = None
            Position of plane in x direction, only one of x,y,z can be specified to define plane.
        y : float = None
            Position of plane in y direction, only one of x,y,z can be specified to define plane.
        z : float = None
            Position of plane in z direction, only one of x,y,z can be specified to define plane.
        permittivity_threshold : float = 1
            Permittivity value used to define the shape boundaries for structures with custom
            medim
        frequency : float = 0
            Frequency for permittivity evaluation in case of custom medium (Hz).
        gds_layer_dtype_map : Dict
            Dictionary mapping mediums to GDSII layer and data type tuples.
        Return
        ------
        List
            List of `gdstk.Polygon`.
        """
        sim, _ = self.to_simulation()
        return sim.to_gdstk(
            x=x,
            y=y,
            z=z,
            permittivity_threshold=permittivity_threshold,
            frequency=frequency,
            gds_layer_dtype_map=gds_layer_dtype_map,
        )

    def plot(
        self,
        x: Optional[float] = None,
        y: Optional[float] = None,
        z: Optional[float] = None,
        ax: Ax = None,
        source_alpha: Optional[float] = None,
        monitor_alpha: Optional[float] = None,
        hlim: Optional[tuple[float, float]] = None,
        vlim: Optional[tuple[float, float]] = None,
        **patch_kwargs,
    ) -> Ax:
        """Wrapper around regular :class:`.Simulation` structure plotting."""
        sim, _ = self.to_simulation()
        return sim.plot(
            x=x,
            y=y,
            z=z,
            ax=ax,
            source_alpha=source_alpha,
            monitor_alpha=monitor_alpha,
            hlim=hlim,
            vlim=vlim,
            **patch_kwargs,
        )

    def plot_eps(
        self,
        x: Optional[float] = None,
        y: Optional[float] = None,
        z: Optional[float] = None,
        freq: Optional[float] = None,
        alpha: Optional[float] = None,
        source_alpha: Optional[float] = None,
        monitor_alpha: Optional[float] = None,
        hlim: Optional[tuple[float, float]] = None,
        vlim: Optional[tuple[float, float]] = None,
        ax: Ax = None,
    ) -> Ax:
        """Wrapper around regular :class:`.Simulation` permittivity plotting."""
        sim, _ = self.to_simulation()
        return sim.plot_eps(
            x=x,
            y=y,
            z=z,
            ax=ax,
            source_alpha=source_alpha,
            monitor_alpha=monitor_alpha,
            hlim=hlim,
            vlim=vlim,
        )

    def plot_structures(
        self,
        x: Optional[float] = None,
        y: Optional[float] = None,
        z: Optional[float] = None,
        ax: Ax = None,
        hlim: Optional[tuple[float, float]] = None,
        vlim: Optional[tuple[float, float]] = None,
    ) -> Ax:
        """Plot each of simulation's structures on a plane defined by one nonzero x,y,z coordinate.

        Parameters
        ----------
        x : float = None
            position of plane in x direction, only one of x, y, z must be specified to define plane.
        y : float = None
            position of plane in y direction, only one of x, y, z must be specified to define plane.
        z : float = None
            position of plane in z direction, only one of x, y, z must be specified to define plane.
        ax : matplotlib.axes._subplots.Axes = None
            Matplotlib axes to plot on, if not specified, one is created.
        hlim : Tuple[float, float] = None
            The x range if plotting on xy or xz planes, y range if plotting on yz plane.
        vlim : Tuple[float, float] = None
            The z range if plotting on xz or yz planes, y plane if plotting on xy plane.

        Returns
        -------
        matplotlib.axes._subplots.Axes
            The supplied or created matplotlib axes.
        """
        sim, _ = self.to_simulation()
        return sim.plot_structures(
            x=x,
            y=y,
            z=z,
            ax=ax,
            hlim=hlim,
            vlim=vlim,
        )

    def plot_structures_eps(
        self,
        x: Optional[float] = None,
        y: Optional[float] = None,
        z: Optional[float] = None,
        freq: Optional[float] = None,
        alpha: Optional[float] = None,
        cbar: bool = True,
        reverse: bool = False,
        ax: Ax = None,
        hlim: Optional[tuple[float, float]] = None,
        vlim: Optional[tuple[float, float]] = None,
    ) -> Ax:
        """Plot each of simulation's structures on a plane defined by one nonzero x,y,z coordinate.
        The permittivity is plotted in grayscale based on its value at the specified frequency.

        Parameters
        ----------
        x : float = None
            position of plane in x direction, only one of x, y, z must be specified to define plane.
        y : float = None
            position of plane in y direction, only one of x, y, z must be specified to define plane.
        z : float = None
            position of plane in z direction, only one of x, y, z must be specified to define plane.
        freq : float = None
            Frequency to evaluate the relative permittivity of all mediums.
            If not specified, evaluates at infinite frequency.
        reverse : bool = False
            If ``False``, the highest permittivity is plotted in black.
            If ``True``, it is plotteed in white (suitable for black backgrounds).
        cbar : bool = True
            Whether to plot a colorbar for the relative permittivity.
        alpha : float = None
            Opacity of the structures being plotted.
            Defaults to the structure default alpha.
        ax : matplotlib.axes._subplots.Axes = None
            Matplotlib axes to plot on, if not specified, one is created.
        hlim : Tuple[float, float] = None
            The x range if plotting on xy or xz planes, y range if plotting on yz plane.
        vlim : Tuple[float, float] = None
            The z range if plotting on xz or yz planes, y plane if plotting on xy plane.

        Returns
        -------
        matplotlib.axes._subplots.Axes
            The supplied or created matplotlib axes.
        """
        sim, _ = self.to_simulation()
        return sim.plot_structures_eps(
            x=x,
            y=y,
            z=z,
            freq=freq,
            alpha=alpha,
            cbar=cbar,
            reverse=reverse,
            ax=ax,
            hlim=hlim,
            vlim=vlim,
        )

    def epsilon(
        self,
        box: Box,
        coord_key: str = "centers",
        freq: Optional[float] = None,
    ) -> xr.DataArray:
        """Get array of permittivity at volume specified by box and freq.

        Parameters
        ----------
        box : :class:`.Box`
            Rectangular geometry specifying where to measure the permittivity.
        coord_key : str = 'centers'
            Specifies at what part of the grid to return the permittivity at.
            Accepted values are ``{'centers', 'boundaries', 'Ex', 'Ey', 'Ez', 'Exy', 'Exz', 'Eyx',
            'Eyz', 'Ezx', Ezy'}``. The field values (eg. ``'Ex'``) correspond to the corresponding field
            locations on the yee lattice. If field values are selected, the corresponding diagonal
            (eg. ``eps_xx`` in case of ``'Ex'``) or off-diagonal (eg. ``eps_xy`` in case of ``'Exy'``) epsilon
            component from the epsilon tensor is returned. Otherwise, the average of the main
            values is returned.
        freq : float = None
            The frequency to evaluate the mediums at.
            If not specified, evaluates at infinite frequency.

        Returns
        -------
        xarray.DataArray
            Datastructure containing the relative permittivity values and location coordinates.
            For details on xarray DataArray objects,
            refer to `xarray's Documentation <https://tinyurl.com/2zrzsp7b>`_.

        See Also
        --------

        **Notebooks**
            * `First walkthrough: permittivity data <../../notebooks/Simulation.html#Permittivity-data>`_
        """
        sim, _ = self.to_simulation()
        return sim.epsilon(box=box, coord_key=coord_key, freq=freq)

    def __eq__(self, other: JaxSimulation) -> bool:
        """Are two JaxSimulation objects equal?"""
        return self.to_simulation()[0] == other.to_simulation()[0]

    @classmethod
    def split_monitors(cls, monitors: list[Monitor], jax_info: JaxInfo) -> dict[str, Monitor]:
        """Split monitors into user and adjoint required based on jax info."""

        all_monitors = list(monitors)

        # grab or compute the number of type of monitor
        num_grad_monitors = jax_info.num_grad_monitors
        num_grad_eps_monitors = jax_info.num_grad_eps_monitors
        num_output_monitors = jax_info.num_output_monitors
        num_total_monitors = len(all_monitors)
        num_mnts = (
            num_total_monitors - num_grad_monitors - num_output_monitors - num_grad_eps_monitors
        )

        # split the monitor list based on these numbers
        monitors = all_monitors[:num_mnts]
        output_monitors = all_monitors[num_mnts : num_mnts + num_output_monitors]
        grad_monitors = all_monitors[
            num_mnts + num_output_monitors : num_mnts + num_output_monitors + num_grad_monitors
        ]
        grad_eps_monitors = all_monitors[num_mnts + num_output_monitors + num_grad_monitors :]

        # load into a dictionary
        return {
            "monitors": monitors,
            "output_monitors": output_monitors,
            "grad_monitors": grad_monitors,
            "grad_eps_monitors": grad_eps_monitors,
        }

    @classmethod
    def split_structures(
        cls, structures: list[Structure], jax_info: JaxInfo
    ) -> dict[str, Structure]:
        """Split structures into regular and input based on jax info."""

        all_structures = list(structures)

        # get numbers of regular and input structures
        num_input_structures = jax_info.num_input_structures
        num_structs = len(structures) - num_input_structures

        # split the list based on these numbers
        structures = all_structures[:num_structs]
        structure_type_map = {
            "JaxStructure": JaxStructure,
            "JaxStructureStaticMedium": JaxStructureStaticMedium,
            "JaxStructureStaticGeometry": JaxStructureStaticGeometry,
        }

        input_structures = []
        for struct_type_str, struct in zip(
            jax_info.input_structure_types, all_structures[num_structs:]
        ):
            struct_type = structure_type_map[struct_type_str]
            new_structure = struct_type.from_structure(struct)
            input_structures.append(new_structure)

        # return a dictionary containing these split structures
        return {"structures": structures, "input_structures": input_structures}

    @classmethod
    def from_simulation(cls, simulation: Simulation, jax_info: JaxInfo) -> JaxSimulation:
        """Convert :class:`.Simulation` to :class:`.JaxSimulation` with extra info."""

        sim_dict = simulation.dict(exclude={"type", "structures", "monitors"})

        # split structures and monitors into their respective fields for JaxSimulation
        structures = cls.split_structures(structures=simulation.structures, jax_info=jax_info)
        monitors = cls.split_monitors(monitors=simulation.monitors, jax_info=jax_info)

        # update the dictionary with these and the adjoint fwidth
        sim_dict.update(**structures)
        sim_dict.update(**monitors)
        sim_dict.update(
            {
                "fwidth_adjoint": jax_info.fwidth_adjoint,
                "run_time_adjoint": jax_info.run_time_adjoint,
            }
        )

        # load JaxSimulation from the dictionary
        return cls.parse_obj(sim_dict)

    @classmethod
    def make_sim_fwd(cls, simulation: Simulation, jax_info: JaxInfo) -> tuple[Simulation, JaxInfo]:
        """Make the forward :class:`.JaxSimulation` from the supplied :class:`.Simulation`."""

        mnt_dict = JaxSimulation.split_monitors(monitors=simulation.monitors, jax_info=jax_info)
        structure_dict = JaxSimulation.split_structures(
            structures=simulation.structures, jax_info=jax_info
        )
        output_monitors = mnt_dict["output_monitors"]
        input_structures = structure_dict["input_structures"]
        grad_mnt_dict = cls.get_grad_monitors(
            input_structures=input_structures,
            freqs_adjoint=cls.get_freqs_adjoint(output_monitors=output_monitors),
        )

        grad_mnts = grad_mnt_dict["grad_monitors"]
        grad_eps_mnts = grad_mnt_dict["grad_eps_monitors"]

        full_monitors = list(simulation.monitors) + grad_mnts + grad_eps_mnts

        # jax_sim_fwd = jax_sim.updated_copy(**grad_mnts)
        # sim_fwd, jax_info = jax_sim_fwd.to_simulation()

        sim_fwd = simulation.updated_copy(monitors=full_monitors)
        jax_info = jax_info.updated_copy(
            num_grad_monitors=len(grad_mnts),
            num_grad_eps_monitors=len(grad_eps_mnts),
        )

        # cls.split_monitors(monitors=simulation.monitors, jax_info=jax_info)
        # sim_fwd = simulation.updated_copy()

        return sim_fwd, jax_info

    def to_simulation_fwd(self) -> tuple[Simulation, JaxInfo, JaxInfo]:
        """Like ``to_simulation()`` but the gradient monitors are included."""
        simulation, jax_info = self.to_simulation()
        sim_fwd, jax_info_fwd = self.make_sim_fwd(simulation=simulation, jax_info=jax_info)
        return sim_fwd, jax_info_fwd, jax_info

    @staticmethod
    def get_grad_monitors(
        input_structures: list[Structure], freqs_adjoint: list[float], include_eps_mnts: bool = True
    ) -> dict:
        """Return dictionary of gradient monitors for simulation."""
        grad_mnts = []
        grad_eps_mnts = []
        for index, structure in enumerate(input_structures):
            grad_mnt, grad_eps_mnt = structure.make_grad_monitors(
                freqs=freqs_adjoint, name=f"grad_mnt_{index}"
            )
            grad_mnts.append(grad_mnt)
            if include_eps_mnts:
                grad_eps_mnts.append(grad_eps_mnt)
        return {"grad_monitors": grad_mnts, "grad_eps_monitors": grad_eps_mnts}

    def _store_vjp_structure(
        self,
        structure: JaxStructure,
        fld_fwd: FieldData,
        fld_adj: FieldData,
        eps_data: PermittivityData,
        num_proc: int = NUM_PROC_LOCAL,
    ) -> JaxStructure:
        """Store the vjp for a single structure."""

        freq_max = float(max(eps_data.eps_xx.coords["f"]))
        eps_out = self.medium.eps_model(frequency=freq_max)
        return structure.store_vjp(
            grad_data_fwd=fld_fwd,
            grad_data_adj=fld_adj,
            grad_data_eps=eps_data,
            sim_bounds=self.bounds,
            eps_out=eps_out,
            num_proc=num_proc,
        )

    def store_vjp(
        self,
        grad_data_fwd: tuple[FieldData],
        grad_data_adj: tuple[FieldData],
        grad_eps_data: tuple[PermittivityData],
        num_proc: int = NUM_PROC_LOCAL,
    ) -> JaxSimulation:
        """Store the vjp w.r.t. each input_structure as a sim using fwd and adj grad_data."""

        # if num_proc supplied and greater than 1, run parallel
        if num_proc is not None and num_proc > 1:
            return self.store_vjp_parallel(
                grad_data_fwd=grad_data_fwd,
                grad_data_adj=grad_data_adj,
                grad_eps_data=grad_eps_data,
                num_proc=num_proc,
            )

        # otherwise, call regular sequential one
        return self.store_vjp_sequential(
            grad_data_fwd=grad_data_fwd, grad_data_adj=grad_data_adj, grad_eps_data=grad_eps_data
        )

    def store_vjp_sequential(
        self,
        grad_data_fwd: tuple[FieldData],
        grad_data_adj: tuple[FieldData],
        grad_eps_data: tuple[PermittivityData],
    ) -> JaxSimulation:
        """Store the vjp w.r.t. each input_structure without multiprocessing."""
        map_args = [self.input_structures, grad_data_fwd, grad_data_adj, grad_eps_data]
        input_structures_vjp = list(map(self._store_vjp_structure, *map_args))

        return self.copy(
            update={
                "input_structures": input_structures_vjp,
                "grad_monitors": (),
                "grad_eps_monitors": (),
            }
        )

    def store_vjp_parallel(
        self,
        grad_data_fwd: tuple[FieldData],
        grad_data_adj: tuple[FieldData],
        grad_eps_data: tuple[PermittivityData],
        num_proc: int,
    ) -> JaxSimulation:
        """Store the vjp w.r.t. each input_structure as a sim using fwd and adj grad_data, and
        parallel processing over ``num_proc`` processes."""

        # Indexing into structures which use internal parallelization, and those which don't.
        # For the latter, simple parallelization over the list will be used.
        internal_par_structs = [JaxGeometryGroup]

        # Parallelize polyslabs internally or externally depending on total number
        polyslabs = [struct for struct in self.input_structures if isinstance(struct, JaxPolySlab)]
        if len(polyslabs) < num_proc:
            internal_par_structs += [JaxPolySlab]

        inds_par_internal, inds_par_external = [], []
        for index, structure in enumerate(self.input_structures):
            if isinstance(structure.geometry, tuple(internal_par_structs)):
                inds_par_internal.append(index)
            else:
                inds_par_external.append(index)

        def make_args(indexes, num_proc_internal) -> list:
            """Make the arguments to map over selecting over a set of structure ``indexes``."""
            args_list = []
            for index in indexes:
                args_i = [
                    self.input_structures[index],
                    grad_data_fwd[index],
                    grad_data_adj[index],
                    grad_eps_data[index],
                    num_proc_internal,
                ]
                args_list.append(args_i)

            return args_list

        # Get vjps for structures that parallelize internally using simple map
        args_list_internal = make_args(inds_par_internal, num_proc_internal=num_proc)
        vjps_par_internal = [self._store_vjp_structure(*args) for args in args_list_internal]

        # Get vjps for structures where we parallelize directly here
        args_list_external = make_args(inds_par_external, num_proc_internal=NUM_PROC_LOCAL)

        vjps_par_external = list(
            Parallel(n_jobs=num_proc)(
                delayed(self._store_vjp_structure)(*args) for args in args_list_external
            )
        )

        # Reshuffle the two lists back in the correct order
        vjps_all = list(vjps_par_internal) + list(vjps_par_external)
        input_structures_vjp = [None] * len(self.input_structures)
        for index, vjp in zip(inds_par_internal + inds_par_external, vjps_all):
            input_structures_vjp[index] = vjp

        return self.copy(
            update={
                "input_structures": input_structures_vjp,
                "grad_monitors": (),
                "grad_eps_monitors": (),
            }
        )



from __future__ import annotations

import json

from tidy3d.components.base import Tidy3dBaseModel
from tidy3d.components.data.index import SimulationDataMap
from tidy3d.log import log
from tidy3d.plugins.smatrix.component_modelers.modal import ModalComponentModeler
from tidy3d.plugins.smatrix.component_modelers.terminal import TerminalComponentModeler
from tidy3d.plugins.smatrix.component_modelers.types import ComponentModelerType
from tidy3d.plugins.smatrix.data.modal import ModalComponentModelerData
from tidy3d.plugins.smatrix.data.terminal import TerminalComponentModelerData
from tidy3d.plugins.smatrix.data.types import ComponentModelerDataType
from tidy3d.web import Batch, BatchData

DEFAULT_DATA_DIR = "."


def compose_modeler(
    modeler_file: str,
) -> ComponentModelerType:
    """Load a component modeler from an HDF5 file.

    This function reads an HDF5 file, determines the modeler type
    (`ModalComponentModeler` or `TerminalComponentModeler`), and constructs the
    corresponding modeler object.

    Parameters
    ----------
    modeler_file : str
        Path to the HDF5 file containing the modeler definition.

    Returns
    -------
    ComponentModelerType
        The loaded `ModalComponentModeler` or `TerminalComponentModeler` object.

    Raises
    ------
    TypeError
        If the modeler type specified in the file is not supported.
    """
    json_str = Tidy3dBaseModel._json_string_from_hdf5(modeler_file)
    model_dict = json.loads(json_str)
    modeler_type = model_dict["type"]

    if modeler_type == "ModalComponentModeler":
        modeler = ModalComponentModeler.from_file(modeler_file)
    elif modeler_type == "TerminalComponentModeler":
        modeler = TerminalComponentModeler.from_file(modeler_file)
    else:
        raise TypeError(f"Unsupported modeler type: {type(modeler_type).__name__}")
    return modeler


def compose_modeler_data(
    modeler: ModalComponentModeler | TerminalComponentModeler,
    indexed_sim_data: SimulationDataMap,
) -> ComponentModelerDataType:
    """Create a modeler data object from a modeler and indexed simulation data.

    This function acts as a dispatcher, creating either a
    `ModalComponentModelerData` or `TerminalComponentModelerData` object based on
    the type of the input `modeler`.

    Parameters
    ----------
    modeler : ModalComponentModeler | TerminalComponentModeler
        The component modeler for which to create the data object.
    indexed_sim_data : SimulationDataMap
        A map of simulation data indexed by port names.

    Returns
    -------
    ComponentModelerDataType
        The appropriate data object containing the simulation results.

    Raises
    ------
    TypeError
        If the provided `modeler` is not a recognized type.
    """
    if isinstance(modeler, ModalComponentModeler):
        modeler_data = ModalComponentModelerData(modeler=modeler, data=indexed_sim_data)
    elif isinstance(modeler, TerminalComponentModeler):
        modeler_data = TerminalComponentModelerData(modeler=modeler, data=indexed_sim_data)
    else:
        raise TypeError(f"Unsupported modeler type: {type(modeler).__name__}")
    return modeler_data


def compose_modeler_data_from_batch_data(
    modeler: ComponentModelerType,
    batch_data: BatchData,
) -> ComponentModelerDataType:
    """Select the correct composer based on modeler type and create the data object.

    This method acts as a dispatcher, inspecting the type of `modeler` to determine
    which composer function to invoke. It populates a `ComponentModelerData`
    object with results from a `BatchData` object.

    Parameters
    ----------
    modeler : ComponentModelerType
        The component modeler, which can be either a `ModalComponentModeler` or
        a `TerminalComponentModeler`.
    batch_data : BatchData
        The results obtained from running the simulation `Batch`.

    Returns
    -------
    ComponentModelerDataType
        The appropriate data object containing the simulation results.

    Raises
    ------
    TypeError
        If the provided `modeler` is not a recognized type.
    """
    port_simulation_data = SimulationDataMap(
        keys=tuple(batch_data.keys()), values=tuple(batch_data.values())
    )
    return compose_modeler_data(modeler=modeler, indexed_sim_data=port_simulation_data)


def create_batch(
    modeler: ComponentModelerType,
    **kwargs,
) -> Batch:
    """Create a simulation Batch from a component modeler.

    Parameters
    ----------
    modeler : ComponentModelerType
        The component modeler that defines the set of simulations.
    **kwargs
        Additional keyword arguments passed to the `Batch` constructor.

    Returns
    -------
    Batch
        The configured `Batch` object ready for execution.
    """

    batch = Batch(
        simulations=modeler.sim_dict,
        **kwargs,
    )
    return batch


def _run_local(
    modeler: ComponentModelerType,
    path_dir: str = DEFAULT_DATA_DIR,
    **kwargs,
) -> ComponentModelerDataType:
    """Execute the full simulation workflow for a given component modeler.

    This function orchestrates the end-to-end process:
    1. Creates a `Batch` of simulations from the `modeler`.
    2. Submits the `Batch` for execution and waits for results.
    3. Composes the results into a structured `ComponentModelerDataType` object.

    Parameters
    ----------
    modeler : ComponentModelerType
        The component modeler defining the simulations to be run.
    path_dir : str, optional
        The directory where the batch file will be saved. Defaults to ".".
    **kwargs
        Extra keyword arguments propagated to the Batch creation.

    Returns
    -------
    ComponentModelerDataType
        An object containing the processed simulation data, ready for
        S-parameter extraction and analysis.
    """

    # autograd path if any sim is valid for autograd
    from tidy3d.web.api.autograd import autograd as web_ag

    sims = modeler.sim_dict
    if any(web_ag.is_valid_for_autograd(sim) for sim in sims.values()):
        if len(modeler.element_mappings) > 0:
            log.warning(
                "Element mappings are used to populate S-matrix values, but autograd gradients "
                "are computed only for simulated elements. Gradients for mapped elements are not "
                "included. For optimization with autograd, prefer enforcing symmetry in geometry/"
                "objective functions and use 'run_only' to select unique sources.",
                log_once=True,
            )

        from tidy3d.web.api.autograd.autograd import _run_async

        kwargs.setdefault("folder_name", "default")
        kwargs.setdefault("simulation_type", "tidy3d_autograd_async")
        kwargs.setdefault("path_dir", path_dir)

        sim_data_map = _run_async(simulations=sims, **kwargs)

        return compose_modeler_data_from_batch_data(modeler=modeler, batch_data=sim_data_map)

    batch = create_batch(modeler=modeler)
    priority = kwargs.get("priority")
    if priority is None:
        batch_data = batch.run(path_dir=path_dir)
    else:
        batch_data = batch.run(path_dir=path_dir, priority=priority)
    return compose_modeler_data_from_batch_data(modeler=modeler, batch_data=batch_data)



from __future__ import annotations

from collections import deque
from typing import Callable, Literal, Optional, Union

import autograd.numpy as np
import pydantic.v1 as pd
from autograd import value_and_grad
from numpy.typing import NDArray
from scipy.optimize import minimize

import tidy3d as td
from tidy3d.components.base import Tidy3dBaseModel
from tidy3d.components.grid.grid import Coords
from tidy3d.plugins.autograd.constants import BETA_DEFAULT, ETA_DEFAULT
from tidy3d.plugins.autograd.types import KernelType, PaddingType

from .filters import make_filter
from .projections import tanh_projection


class FilterAndProject(Tidy3dBaseModel):
    """A class that combines filtering and projection operations."""

    radius: Union[float, tuple[float, ...]] = pd.Field(
        ..., title="Radius", description="The radius of the kernel."
    )
    dl: Union[float, tuple[float, ...]] = pd.Field(
        ..., title="Grid Spacing", description="The grid spacing."
    )
    size_px: Union[int, tuple[int, ...]] = pd.Field(
        None, title="Size in Pixels", description="The size of the kernel in pixels."
    )
    beta: pd.NonNegativeFloat = pd.Field(
        BETA_DEFAULT, title="Beta", description="The beta parameter for the tanh projection."
    )
    eta: pd.NonNegativeFloat = pd.Field(
        ETA_DEFAULT, title="Eta", description="The eta parameter for the tanh projection."
    )
    filter_type: KernelType = pd.Field(
        "conic", title="Filter Type", description="The type of filter to create."
    )
    padding: PaddingType = pd.Field(
        "reflect", title="Padding", description="The padding mode to use."
    )

    def __call__(
        self, array: NDArray, beta: Optional[float] = None, eta: Optional[float] = None
    ) -> NDArray:
        """Apply the filter and projection to an input array.

        Parameters
        ----------
        array : np.ndarray
            The input array to filter and project.
        beta : float = None
            The beta parameter for the tanh projection. If None, uses the instance's beta.
        eta : float = None
            The eta parameter for the tanh projection. If None, uses the instance's eta.

        Returns
        -------
        np.ndarray
            The filtered and projected array.
        """
        filter_instance = make_filter(
            radius=self.radius,
            dl=self.dl,
            size_px=self.size_px,
            filter_type=self.filter_type,
            padding=self.padding,
        )
        filtered = filter_instance(array)
        beta = beta if beta is not None else self.beta
        eta = eta if eta is not None else self.eta
        projected = tanh_projection(filtered, beta, eta)
        return projected


def make_filter_and_project(
    radius: Optional[Union[float, tuple[float, ...]]] = None,
    dl: Optional[Union[float, tuple[float, ...]]] = None,
    *,
    size_px: Optional[Union[int, tuple[int, ...]]] = None,
    beta: float = BETA_DEFAULT,
    eta: float = ETA_DEFAULT,
    filter_type: KernelType = "conic",
    padding: PaddingType = "reflect",
) -> Callable:
    """Create a function that filters and projects an array.

    See Also
    --------
    :func:`~parametrizations.FilterAndProject`.
    """
    return FilterAndProject(
        radius=radius,
        dl=dl,
        size_px=size_px,
        beta=beta,
        eta=eta,
        filter_type=filter_type,
        padding=padding,
    )


def initialize_params_from_simulation(
    sim: td.Simulation,
    param_to_structure: Callable[..., td.Structure],
    params0: np.ndarray,
    *,
    freq: Optional[float] = None,
    outside_handling: Literal["extrapolate", "mask", "nan"] = "mask",
    maxiter: int = 100,
    bounds: tuple[Optional[float], Optional[float]] = (0.0, 1.0),
    rel_improve_tol: float = 1e-3,
    verbose: bool = False,
    **param_kwargs,
) -> np.ndarray:
    """Initialize design parameters to match base simulation permittivity in a region.

    Builds an objective that compares the base simulation's permittivity against
    the permittivity generated by a user‑supplied
    parameterization and minimizes the relative L2 error using L‑BFGS‑B.

    Notes
    -----
    - The simulation is not modified. If you need finer sampling inside the design
      region, add a :class:`.MeshOverrideStructure` to ``sim.grid_spec`` first.
    - The callable ``param_to_structure`` controls the parameterization and must return
      a :class:`.Structure` (typically with a :class:`.CustomMedium`). Its permittivity
      dataset (and coordinates) defines the grid used for comparison.
    - The base simulation permittivity is sampled once on an extended subgrid covering
      the design geometry and then interpolated onto the design region coordinates
      using coordinate-aware interpolation (no per-iteration interpolation).
    - Early stopping uses a single knob ``rel_improve_tol``; optimization stops once the
      relative improvement over a small fixed window falls below this value.
    - Points outside the base‑epsilon coverage can be handled via ``outside_handling``:
      ``'extrapolate'`` (use nearest extrapolation, default in earlier versions),
      ``'mask'`` (ignore outside points using a coverage mask; default), or
      ``'nan'`` (non‑extended grid; treat outside as NaN and ignore in the loss).

    Parameters
    ----------
    sim : :class:`.Simulation`
        Base simulation without the design structure.
    param_to_structure : Callable[..., :class:`.Structure`]
        Function mapping parameters to a :class:`.Structure` whose medium permittivity is used
        for comparison. Any extra keyword arguments passed to this initializer are forwarded to
        ``param_to_structure``.
    params0 : :class:`numpy.ndarray`
        Initial parameter array (2D or 3D). Values may take any range, consistent with ``bounds``
        and the expectations of ``param_to_structure``.
    maxiter : int = 100
        Maximum number of L‑BFGS‑B iterations.
    freq : float, optional
        Frequency at which permittivity is evaluated. If ``None``, uses infinite frequency.
    outside_handling : {"extrapolate", "mask", "nan"} = "mask"
        Strategy for points where design coordinates fall outside the sampled base epsilon:
        - "extrapolate": include them using nearest extrapolation.
        - "mask": include only points within the coverage bounds of ``sim.epsilon`` on the
          extended subgrid.
        - "nan": sample on a non‑extended subgrid and ignore points where interpolation returns NaN.
    bounds : tuple[float | None, float | None] = (0.0, 1.0)
        Element‑wise parameter bounds, e.g. ``(0.0, 1.0)`` or ``(-1.0, 1.0)``. Use ``None`` to
        indicate an unbounded side.
    rel_improve_tol : float = 1e-3
        Early‑stop tolerance on relative improvement over a small window.
    verbose : bool = False
        If ``True``, prints SciPy optimizer messages.

    Returns
    -------
    :class:`numpy.ndarray`
        Optimized parameters with the same shape as ``params0``.

    Examples
    --------
    Initialize a small parameter array in a 2D simulation using a minimal,
    differentiable parameterization that maps parameters to a permittivity field.

    >>> import autograd.numpy as np
    >>> import tidy3d as td
    >>> from tidy3d.plugins.autograd.invdes import initialize_params_from_simulation
    >>> sim = td.Simulation(
    ...     size=(2.0, 2.0, 0.0),
    ...     grid_spec=td.GridSpec.uniform(dl=1.0),
    ...     boundary_spec=td.BoundarySpec.pml(x=True, y=True),
    ...     run_time=1e-12,
    ... )
    >>> box = td.Box(center=(0, 0, 0), size=(1.0, 1.0, 1.0))
    >>> params0 = np.zeros((3, 3))
    >>> def param_to_structure(p):
    ...     # Map params -> density in [0, 1], then to eps in [1, 4].
    ...     density = 0.5 * (1 + np.tanh(p))
    ...     eps = 1.0 + 3.0 * density
    ...     eps3d = eps.reshape((*p.shape, 1))
    ...     return td.Structure.from_permittivity_array(geometry=box, eps_data=eps3d)
    >>> params = initialize_params_from_simulation(
    ...     sim=sim,
    ...     param_to_structure=param_to_structure,
    ...     params0=params0,
    ...     maxiter=3,
    ... )
    >>> params.shape
    (3, 3)
    """
    structure_init = param_to_structure(params0, **param_kwargs)

    if outside_handling not in ("extrapolate", "mask", "nan"):
        raise ValueError("'outside_handling' must be one of {'extrapolate', 'mask', 'nan'}.")

    # build base epsilon and interpolate onto design coords
    extend_flag = outside_handling != "nan"
    subgrid = sim.discretize(structure_init.geometry, extend=extend_flag)
    eps_base_da = sim.epsilon_on_grid(grid=subgrid, coord_key="centers", freq=freq)

    design_eps_da = structure_init.medium.permittivity
    design_coords = Coords(
        x=np.array(design_eps_da.coords["x"]),
        y=np.array(design_eps_da.coords["y"]),
        z=np.array(design_eps_da.coords["z"]),
    )

    if outside_handling == "nan":
        eps_base_interp = design_coords.spatial_interp(
            array=eps_base_da, interp_method="linear", fill_value=np.nan
        ).data
        mask = np.isfinite(eps_base_interp)
    else:
        eps_base_interp = design_coords.spatial_interp(
            array=eps_base_da, interp_method="linear", fill_value="extrapolate"
        ).data
        if outside_handling == "mask":
            # build mask from coverage bounds of base epsilon coordinates
            xb, yb, zb = [np.array(eps_base_da.coords[d]) for d in ("x", "y", "z")]
            xd, yd, zd = [np.array(design_eps_da.coords[d]) for d in ("x", "y", "z")]
            mask_x = (xd >= np.min(xb)) & (xd <= np.max(xb))
            mask_y = (yd >= np.min(yb)) & (yd <= np.max(yb))
            mask_z = (zd >= np.min(zb)) & (zd <= np.max(zb))
            mask = (mask_x[:, None, None]) & (mask_y[None, :, None]) & (mask_z[None, None, :])
        else:
            mask = None

    if mask is not None:
        covered = int(np.sum(mask))
        total = int(np.prod(mask.shape))
        frac = covered / max(total, 1)
        if frac < 0.9:
            td.log.warning(
                f"Only {frac:.1%} of design points are covered by base epsilon sampling. "
                "Consider adding a 'MeshOverrideStructure' or adjusting design coordinates."
            )

    if mask is None:
        denom = np.sqrt(np.sum(eps_base_interp.real**2 + eps_base_interp.imag**2))
    else:
        denom = np.sqrt(
            np.sum((eps_base_interp.real[mask]) ** 2 + (eps_base_interp.imag[mask]) ** 2)
        )
    denom = 1.0 if denom == 0 else denom

    def loss_fn(params_vec: np.ndarray) -> float:
        params = params_vec.reshape(params0.shape)
        structure = param_to_structure(params, **param_kwargs)
        eps_design = structure.medium.permittivity.data
        if mask is None:
            res = eps_base_interp - eps_design
            return 0.5 * np.sum(res.real**2 + res.imag**2) / denom
        res = (eps_base_interp - eps_design)[mask]
        return 0.5 * np.sum(res.real**2 + res.imag**2) / denom

    val_and_grad = value_and_grad(loss_fn)

    WINDOW = 3
    state = {
        "vals": deque(maxlen=WINDOW),
        "best_val": np.inf,
        "best_x": params0.ravel().copy(),
    }

    def callback(xk: np.ndarray):
        val = loss_fn(xk)
        if val < state["best_val"]:
            state["best_val"] = val
            state["best_x"] = xk.copy()
        state["vals"].append(val)
        if len(state["vals"]) == WINDOW:
            v_then = state["vals"][0]
            v_now = state["vals"][-1]
            rel_improve = (v_then - v_now) / max(abs(v_then), 1e-12)
            if rel_improve < rel_improve_tol:
                raise StopIteration

    bounds_list = [bounds] * params0.size
    try:
        res = minimize(
            fun=val_and_grad,
            x0=params0.ravel(),
            method="L-BFGS-B",
            jac=True,
            bounds=bounds_list,
            callback=callback,
            options={"maxiter": maxiter, "disp": verbose},
        )
        x_final = res.x
    except StopIteration:
        x_final = state["best_x"]
    return x_final.reshape(params0.shape)



from __future__ import annotations

from typing import Union

from tidy3d.components.types.simulation import SimulationDataType, SimulationType
from tidy3d.plugins.smatrix.component_modelers.modal import (
    ModalComponentModeler,
)
from tidy3d.plugins.smatrix.component_modelers.terminal import (
    TerminalComponentModeler,
)
from tidy3d.plugins.smatrix.data.modal import (
    ModalComponentModelerData,
)
from tidy3d.plugins.smatrix.data.terminal import (
    TerminalComponentModelerData,
)

WorkflowType = Union[
    SimulationType,
    ModalComponentModeler,
    TerminalComponentModeler,
]
WorkflowDataType = Union[
    SimulationDataType,
    ModalComponentModelerData,
    TerminalComponentModelerData,
]



"""Utility functions for S-matrix calculations and conversions.

This module provides helper functions for scattering matrix computations,
impedance conversions, and wave amplitude calculations in electromagnetic
simulations.
"""

from __future__ import annotations

from typing import Union

import numpy as np

from tidy3d.components.data.data_array import (
    DataArray,
    FreqDataArray,
)
from tidy3d.components.data.sim_data import SimulationData
from tidy3d.components.types import ArrayFloat1D
from tidy3d.exceptions import Tidy3dError
from tidy3d.plugins.smatrix.data.data_array import PortDataArray, TerminalPortDataArray
from tidy3d.plugins.smatrix.ports.types import (
    LumpedPortType,
    PortCurrentType,
    PortVoltageType,
    TerminalPortType,
)
from tidy3d.plugins.smatrix.types import SParamDef


def port_array_inv(matrix: DataArray):
    """Helper to invert a port matrix.

    Parameters
    ----------
    matrix : DataArray
        The matrix to invert.

    Returns
    -------
    np.ndarray
        The inverted matrix.
    """
    return np.linalg.inv(matrix)


def ab_to_s(
    a_matrix: TerminalPortDataArray, b_matrix: TerminalPortDataArray
) -> TerminalPortDataArray:
    """Get the scattering matrix given the wave amplitude matrices.

    Parameters
    ----------
    a_matrix : TerminalPortDataArray
        Matrix of incident power wave amplitudes.
    b_matrix : TerminalPortDataArray
        Matrix of reflected power wave amplitudes.

    Returns
    -------
    TerminalPortDataArray
        The computed scattering (S) matrix.
    """
    validate_square_matrix(a_matrix, "ab_to_s")
    # Ensure dimensions are ordered properly
    a_matrix = a_matrix.transpose(*TerminalPortDataArray._dims)
    b_matrix = b_matrix.transpose(*TerminalPortDataArray._dims)

    s_matrix = a_matrix.copy(deep=True)
    a_vals = s_matrix.copy(deep=True).values
    b_vals = b_matrix.copy(deep=True).values

    s_vals = np.matmul(b_vals, port_array_inv(a_vals))

    s_matrix.data = s_vals
    return s_matrix


def check_port_impedance_sign(Z_numpy: np.ndarray):
    """Sanity check for consistent sign of real part of Z for each port.

    This check iterates through each port and ensures that the sign of the real
    part of its impedance does not change across all frequencies. A sign change
    can indicate an unphysical result or numerical instability.

    Parameters
    ----------
    Z_numpy : np.ndarray
        NumPy array of impedance values with shape (num_freqs, num_ports).

    Raises
    ------
    Tidy3dError
        If an inconsistent sign of the real part of the impedance
        is detected for any port.
    """
    for port_idx in range(Z_numpy.shape[1]):
        port_Z = Z_numpy[:, port_idx]
        signs = np.sign(np.real(port_Z))
        if not np.all(signs == signs[0]):
            raise Tidy3dError(
                f"Inconsistent sign of real part of Z detected for port {port_idx}. "
                "If you received this error, please create an issue in the Tidy3D "
                "github repository."
            )


def compute_F(Z_numpy: ArrayFloat1D, s_param_def: SParamDef = "pseudo"):
    r"""Helper to convert port impedance matrix to F, which is used for
    computing scattering parameters

    The matrix F is used when converting between S and Z parameters for circuits
    with differing port impedances. Its diagonal elements are defined as

    .. math::

        F_{kk} = 1 / (2 * \sqrt{Re(Z_k)})


    Parameters
    ----------
    Z_numpy : ArrayFloat1D
        NumPy array of complex port impedances.
    s_param_def : SParamDef, optional
        The type of wave amplitudes, by default "pseudo".

    Returns
    -------
    ArrayFloat1D
        NumPy array containing the computed F values.
    """
    # Defined in [2] after equation 4.67
    if s_param_def == "power":
        return 1.0 / (2.0 * np.sqrt(np.real(Z_numpy)))
    # Equation 75 from [1]
    return np.sqrt(np.real(Z_numpy)) / (2.0 * np.abs(Z_numpy))


def compute_port_VI(
    port_out: TerminalPortType, sim_data: SimulationData
) -> tuple[PortVoltageType, PortCurrentType]:
    """Compute the port voltages and currents.

    Parameters
    ----------
    port_out : ``TerminalPortType``
        Port for computing voltage and current.
    sim_data : :class:`.SimulationData`
        Results from simulation containing field data.

    Returns
    -------
    tuple[PortVoltageType, PortCurrentType]
        Voltage and current values at the port as frequency arrays.
    """
    voltage = port_out.compute_voltage(sim_data)
    current = port_out.compute_current(sim_data)
    return voltage, current


def compute_power_wave_amplitudes(
    port: LumpedPortType, sim_data: SimulationData
) -> tuple[FreqDataArray, FreqDataArray]:
    r"""Calculates the unnormalized power wave amplitudes from port voltage (V),
    current (I), and impedance (Z0) using:

    .. math::

        a = (V + Z0*I) / (2 * \sqrt(Re(Z0)))
        b = (V - Z0*I) / (2 * \sqrt(Re(Z0)))

    Parameters
    ----------
    port : :class:`.LumpedPortType`
        Port for computing voltage and current.
    sim_data : :class:`.SimulationData`
        Results from the simulation.

    Returns
    -------
    tuple[FreqDataArray, FreqDataArray]
        Incident (a) and reflected (b) power wave amplitude frequency arrays.
    """
    voltage, current = compute_port_VI(port, sim_data)
    # Amplitudes for the incident and reflected power waves
    a = (voltage + port.impedance * current) / 2 / np.sqrt(np.real(port.impedance))
    b = (voltage - port.impedance * current) / 2 / np.sqrt(np.real(port.impedance))
    return a, b


def compute_power_delivered_by_port(
    port: LumpedPortType, sim_data: SimulationData
) -> FreqDataArray:
    """Compute the power delivered to the network by a lumped port.
    The power is calculated as the incident power minus the reflected power:

    .. math::
        P = 0.5 * (|a|^2 - |b|^2)

    Parameters
    ----------
    port : :class:`.LumpedPortType`
        Port for computing voltage and current.
    sim_data : :class:`.SimulationData`
        Results from the simulation.

    Returns
    -------
    FreqDataArray
        Power in units of Watts as a frequency array.
    """
    a, b = compute_power_wave_amplitudes(port=port, sim_data=sim_data)
    # Power delivered is the incident power minus the reflected power
    return 0.5 * (np.abs(a) ** 2 - np.abs(b) ** 2)


def s_to_z(
    s_matrix: TerminalPortDataArray,
    reference: Union[complex, PortDataArray],
    s_param_def: SParamDef = "pseudo",
) -> DataArray:
    """Get the impedance matrix given the scattering matrix and a reference impedance.

    This function converts an S-matrix to a Z-matrix. It handles both a single
    uniform reference impedance and generalized per-port reference impedances.

    Parameters
    ----------
    s_matrix : :class:`.TerminalPortDataArray`
        Scattering matrix computed using either the pseudo or power wave formulation.
    reference : Union[complex, :class:`.PortDataArray`]
        The reference impedance used at each port.
    s_param_def : SParamDef, optional
        The type of wave amplitudes used for computing the scattering matrix, either pseudo waves
        defined by Equation 53 and Equation 54 in [1] or power waves defined by Equation 4.67 in [2].
        By default "pseudo".

    Returns
    -------
    DataArray
        The computed impedance (Z) matrix.

    Examples
    --------
    The `s_to_z` function is a standalone utility that requires an S-matrix as input.
    This is useful if you have S-matrix data generated
    externally from a :class:`.TerminalComponentModelerData` and want to compare them.

    >>> z_matrix = s_to_z(s_matrix=s_matrix, reference=50, s_param_def="power") # doctest: +SKIP
    >>> z_11 = z_matrix.sel(port_out="port_1", port_in="port_1") # doctest: +SKIP
    """
    validate_square_matrix(s_matrix, "s_to_z")
    # Ensure dimensions are ordered properly
    z_matrix = s_matrix.transpose(*TerminalPortDataArray._dims).copy(deep=True)
    s_vals = z_matrix.values
    eye = np.eye(len(s_matrix.port_out.values), len(s_matrix.port_in.values))[np.newaxis, :, :]
    # Ensure that Zport, F, and Finv act as diagonal matrices when multiplying by left or right
    shape_left = (len(s_matrix.f), len(s_matrix.port_out), 1)
    shape_right = (len(s_matrix.f), 1, len(s_matrix.port_in))
    # Setup the port reference impedance array (scalar)
    if isinstance(reference, PortDataArray):
        Zport = reference.values.reshape(shape_right)
        F = compute_F(Zport, s_param_def).reshape(shape_right)
        Finv = (1.0 / F).reshape(shape_left)
    else:
        Zport = reference
        F = compute_F(Zport, s_param_def)
        Finv = 1.0 / F
    # Use conjugate when S matrix is power-wave based
    if s_param_def == "power":
        Zport_mod = np.conj(Zport)
    else:
        Zport_mod = Zport

    # From equation 74 from [1] for pseudo waves
    # From Equation 4.68 - Pozar - Microwave Engineering 4ed for power waves
    FinvSF = Finv * s_vals * F
    RHS = eye * Zport_mod + FinvSF * Zport
    LHS = eye - FinvSF
    z_vals = np.linalg.solve(LHS, RHS)

    z_matrix.data = z_vals
    return z_matrix


def validate_square_matrix(matrix: TerminalPortDataArray, method_name: str) -> None:
    """Check if the matrix has equal input and output port dimensions.

    Parameters
    ----------
    matrix : TerminalPortDataArray
        Matrix to validate
    method_name : str
        Name of the calling method for error message

    Raises
    ------
    Tidy3dError
        If the matrix is not square (unequal input/output dimensions).
    """
    n_out = len(matrix.port_out)
    n_in = len(matrix.port_in)
    if n_out != n_in:
        raise Tidy3dError(
            f"Cannot compute {method_name}: number of input ports ({n_in}) "
            f"!= the number of output ports ({n_out}). This usually means the 'TerminalComponentModeler' "
            "was run with only a subset of port excitations. Please ensure that the `run_only` field in "
            "the 'TerminalComponentModeler' is not being used."
        )
